/*
 * $RCSfile: riproute.c,v $
 * 
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 *
 */

#include "xkernel.h"
#include "eth.h"
#include "vnet.h"
#include "ip.h"
#include "arp.h"
#include "riproute.h"
#include "rip.h"
#include "rip_i.h"
#include "riproute_i.h"

#ifdef __STDC__

static route *	ript_alloc( RipRouteTable * );
static int	ript_hash(IPhost *);
static route *	ript_new(RipRouteTable *, IPhost *, IPhost *, IPhost *, int, int, int);
static void	ript_timer( Event, void * );
int 		ripHostOnLocalNet( PSTATE *, IPhost * );


#else

static route *	ript_alloc();
static int	ript_hash();
static route *	ript_new();
static void	ript_timer();
int 		ripHostOnLocalNet( PSTATE *, IPhost * );

#endif /* __STDC__ */

static IPhost	ipNull = { 0, 0, 0, 0 };

extern int	traceripp;

xkern_return_t
ript_init( ps, defGw)
    PSTATE *ps;
    IPhost *defGw;
{
    RipRouteTable	*tbl = &ps->rtTbl;
    
    xTrace0(ripp, TR_GROSS_EVENTS, "RIP ript_init()");
    tbl->valid = TRUE;
    tbl->defrt = 0;
    tbl->arr = (route **)xMalloc(ROUTETABLESIZE * sizeof(route *));
    bzero((char *)tbl->arr, ROUTETABLESIZE * sizeof(route *));
    tbl->bpoolsize = BPSIZE;
    if ( IP_EQUAL(*defGw, ipNull) ) {
	xTrace0(ripp, TR_GROSS_EVENTS,
		"RIP routing -- default routing disabled");
    } else {
	if ( ript_add_def(ps, defGw, 0) ) {
	    return XK_FAILURE;
	}
    }
    evDetach( evSchedule( ript_timer, tbl, RTTABLEUPDATE * 1000 ) );
    xTrace0(ripp, TR_GROSS_EVENTS, "RIP ript_init() done");
    return XK_SUCCESS;
}


static route *
ript_alloc( tbl )
    RipRouteTable	*tbl;
{
    route *ptr;
    if ( tbl->bpoolsize == 0 ) {
	xTrace0(ripp, TR_SOFT_ERRORS, "ip ript_alloc ... route table is full");
	return 0;
    }
    tbl->bpoolsize--;
    ptr = (route *)xMalloc(sizeof(route));
    bzero((char *)ptr, sizeof(route));
    return ptr;
}


static route *
ript_new(tbl, net, mask, gw, metric, ttl, ifnum)
    RipRouteTable	*tbl;
    IPhost *net;
    IPhost *mask;
    IPhost *gw;
    int metric;
    int ttl;
    int ifnum;
{
    route *ptr;
    
    ptr = ript_alloc(tbl);
    if ( ptr ) {
	ptr->net = *net;
	ptr->mask = *mask;
	ptr->gw = *gw;
	ptr->metric = metric;
	ptr->ttl = ttl;
	ptr->ifnum = ifnum;
	ptr->next = NULL;
    }
    return ptr;
}


xkern_return_t
ript_add_def( ps, gw, ifnum )
    PSTATE *ps;
    IPhost *gw;
    int	 ifnum;
{
    RipRouteTable	*tbl = &ps->rtTbl;

    xTrace1(ripp, TR_MAJOR_EVENTS,
	    "RIP default route changes.  New GW: %s", ipHostStr(gw));
    if ( ! IP_EQUAL(*gw, ipNull) && ! ripHostOnLocalNet(ps, gw) ) {
	xTrace1(ripp, TR_SOFT_ERRORS,
		"ip: ript_add_def couldn't find interface for gw %s",
		ipHostStr(gw));
	return XK_FAILURE;
    }
    if ( tbl->defrt ) {
	ript_free(tbl, tbl->defrt);
    }
    if ( ! IP_EQUAL(*gw, ipNull) ) {
	tbl->defrt = ript_new( tbl, &ipNull, &ipNull, gw, 1,  0, ifnum );
	if ( tbl->defrt == 0 ) {
	    return XK_FAILURE;
	}
	/*
	 * Re-open the connection for every remote host not connected to the
	 * local net.  This is certainly overkill, but:
	 * 		-- it is not incorrect
	 *		-- the default route shouldn't change often
	 *		-- keeping track of which hosts use the default route
	 *		   is a pain.
	  ipRouteChanged(ps, tbl->defrt, ipRemoteNet);
	 */
    }
    return XK_SUCCESS;
}
  

xkern_return_t
ript_add( pstate, net, mask, gw, metric, ttl, ifnum )
    PSTATE *pstate;
    IPhost *net;
    IPhost *mask;
    IPhost *gw;
    int metric;
    int ttl;
    int ifnum;
{
    route 	*ptr, *srt, *prev;
    u_char  	isdup;
    int  	j;
    u_long 	hashvalue;
    RipRouteTable	*tbl = &pstate->rtTbl;
    
    ptr = ript_new(tbl, net, mask, gw, metric, ttl, ifnum);
    if ( ptr == 0 ) {
	return XK_FAILURE;
    }
    
    /* compute sort key - number of set bits in mask 
       so that route are sorted : host, subnet, net */
    for (j = 0; j < 8; j++)
      ptr->key += ((mask->a >> j) & 1) + 
	((mask->b >> j) & 1) +
	  ((mask->c >> j) & 1) +
	    ((mask->d >> j) & 1) ;
    
    xTrace1(ripp, TR_MORE_EVENTS, "RIP ript_add : key is %d", ptr->key);
    prev = NULL;
    hashvalue = ript_hash(net);
    xTrace1(ripp, TR_MORE_EVENTS, "RIP ript_add : hash value is %d", hashvalue);
    isdup = FALSE;
    for ( srt = tbl->arr[hashvalue]; srt; srt = srt->next ) {
	if ( ptr->key > srt->key )
	  break;
	if ( IP_EQUAL(srt->net, ptr->net) && 
	    IP_EQUAL(srt->mask, ptr->mask) ) {
	    isdup = TRUE;
	    break;
	}
	prev = srt;
    }
    if ( isdup ) {
	route *tmptr;
	if ( IP_EQUAL(srt->gw, ptr->gw) ) {
	    /* update existing route */
	    xTrace0(ripp, TR_MORE_EVENTS, "RIP ript_add: updating existing route");
	    srt->metric = metric;
	    srt->ttl = ttl;
	    srt->ifnum = ifnum;
	    ript_free(tbl, ptr);
	    return XK_SUCCESS;
	}
	/* otherwise someone else has a route there */
	/*
	 * negative metric indicates unconditional override
	 */
	if ( ptr->metric > 0 && srt->metric <= ptr->metric ) {
	    /* it's no better, just drop it */
	    xTrace0(ripp, TR_MORE_EVENTS,
		    "RIP ript_add : dropping duplicate route with greater metric");
	    ript_free(tbl, ptr);
	    return XK_SUCCESS;
	}
	xTrace0(ripp, TR_MORE_EVENTS,
		"RIP ript_add: new duplicate route better, deleting old");
	tmptr = srt;
	srt = srt->next;
	ript_free(tbl, tmptr);
    } else {
	xTrace0(ripp, TR_MORE_EVENTS, "RIP ript_add: adding fresh route");
    }

    /* notify IP here sooner or latter */

    ptr->next = srt;
    if ( prev ) {
	prev->next = ptr;
    } else {
	tbl->arr[hashvalue] = ptr;
    }
    return XK_SUCCESS; 
} /* ript_add */


xkern_return_t
ript_get( tbl, dest, req )
    RipRouteTable	*tbl;
    IPhost 	*dest;
    route	*req;
{
    route *ptr;
    u_long hashvalue;
    u_long sum;
    IPhost fdest;
    
    hashvalue = ript_hash(dest);
    xTrace1(ripp, TR_MORE_EVENTS, "RIP ript_get: hash value is %d",hashvalue);
    for( ptr = tbl->arr[hashvalue]; ptr; ptr = ptr->next ) {
	if ( ptr->ttl == 0 )
	  continue;
	sum =  *(long *)dest & *(long *)&(ptr->mask);
	fdest = *(IPhost *) &sum;
	if ( IP_EQUAL(fdest,ptr->net) )
	  break;
    }
    if ( ptr == 0 ) {
	ptr = tbl->defrt;
    }
    if ( ptr ) {
	*req = *ptr;
	xTrace3(ripp, TR_MORE_EVENTS,
		"RIP ript_get : Mapped host %s to net %s, gw %s",
		ipHostStr(dest), ipHostStr(&ptr->net), ipHostStr(&ptr->gw));
	return XK_SUCCESS;
    } else {
	xTrace1(ripp, TR_SOFT_ERRORS,
		"RIP ript_get: Could not find route for host %s!",
		ipHostStr(dest));
	return XK_FAILURE;
    }
}


void
ript_delete( tbl, net, mask )
    RipRouteTable	*tbl;
    IPhost 	*net, *mask;
{
    route *ptr, *prev;
    u_long  hashvalue; 
    
    hashvalue = ript_hash(net);
    prev = NULL;
    for ( ptr = tbl->arr[hashvalue]; ptr; ptr = ptr->next ) {
	if ( IP_EQUAL(*net, ptr->net) &&
	    IP_EQUAL(*mask, ptr->mask) )
	  break;
	prev = ptr;
    }
    if ( ptr == NULL ) {
	return;
    }
    if ( prev )
      prev->next = ptr->next;
    else
      tbl->arr[hashvalue] = ptr->next;
    ript_free(tbl, ptr);
    return;
}


/* hash value is sum of net portions of RIP address */
static int
ript_hash( net )
     IPhost *net;
{
    IPhost	mask;
    u_long 	hashvalue;
    
    netMaskFind(&mask, net);
    IP_AND(mask, mask, *net);
    hashvalue = mask.a + mask.b + mask.c + mask.d;
    return (hashvalue % ROUTETABLESIZE);
}


void
ript_free(tbl, rt)
    RipRouteTable	*tbl;
    route 	*rt;
{
    tbl->bpoolsize++;
    xFree((char *)rt);
}


static void
ript_timer(ev, arg)
    Event	ev;
    VOID 	*arg;
{
    RipRouteTable	*tbl = (RipRouteTable *)arg;
    route 	*ptr, *prev;
    int 	i;
    
    xTrace0(ripp, TR_EVENTS, "RIP ript_timer called");
    for ( i = 0; i < ROUTETABLESIZE; i++) {
	if ( tbl->arr[i] == 0 ) {
	    continue;
	}
	prev = NULL;
	for ( ptr = tbl->arr[i]; ptr; ) {
	    if ( ptr->ttl != RIPROUTE_TTL_INFINITE ) {
		ptr->ttl -= RTTABLEDELTA;
	    }
	    if ( ptr->ttl == 0 ) {
		if ( prev ) {
		    prev->next = ptr->next;
		    ript_free(tbl, ptr);
		    ptr = prev->next;
		} else {
		    tbl->arr[i] = ptr->next;
		    ript_free(tbl, ptr);
		    ptr = tbl->arr[i];
		}
		continue;
	    }
	    prev = ptr;
	    ptr = ptr->next;
	}
    }
    /*
     * Reschedule this event
     */
    evDetach( evSchedule( ript_timer, tbl, RTTABLEUPDATE * 1000 ) );
}


int
ripHostOnLocalNet( ps, host )
    PSTATE      *ps;
    IPhost      *host;
{
    int         res;
    XObj        llp;

    llp = xGetDown(ps->self, 0);
    xAssert(xIsProtocol(llp));
    res = xControl(llp, VNET_HOSTONLOCALNET, (char *)host, sizeof(IPhost));
    if ( res > 0 ) {
        return 1;
    }
    if ( res < 0 ) {
        xTrace0(ripp, TR_ERRORS, "ip could not do HOSTONLOCALNET call on llp");
    }
    return 0;
}

