/*
 * $RCSfile: tcp_internal.h,v $
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp.h	7.3 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: tcp_internal.h,v $
 * Revision 1.1  1996/07/11 21:40:04  nicko
 * Initial revision
 *
 * Revision 1.40  1995/11/04  00:01:54  ho
 * Added code for dealing with multiple saids on the security participant.
 *
 * Revision 1.39  1995/06/07  21:43:31  spatsch
 * added ipsec support.
 *
 * Revision 1.38  1994/11/22  23:59:17  dorgival
 * Header structure was changed; endianess details now dealt with by usit.
 *
 * Revision 1.38  1994/11/18  22:03:04  dorgival
 * Header structure was changed; endianess details now dealt with by usit.
 *
 * Revision 1.37  1994/04/22  02:15:35  davidm
 * struct in_addr and tcp_seq are now of type xk_u_int32
 * instead of u_long.
 *
 * Revision 1.36  1993/12/16  01:41:38  menze
 * fixed '#else' comment
 *
 */

/**************************************************************************

tcp_internal.h

**************************************************************************/
#include "xkernel.h"
#include "sb.h"
#include "ip.h"
#include "tcp.h"
#include "tcp_port.h"

#ifndef ENDIAN
/*
 * Definitions for byte order,
 * according to byte significance from low address to high.
 */
#define	LITTLE	1234		/* least-significant byte first (vax) */
#define	BIG	4321		/* most-significant byte first */
#define	PDP	3412		/* LSB first in word, MSW first in long (pdp) */
#if defined(vax) || defined(pmax)
#define	ENDIAN	LITTLE
#else
#define	ENDIAN	BIG		/* byte order on mc68000, tahoe, most others */
#endif
#endif /* ENDIAN */

typedef u_long 		n_time;
#undef KPROF
#define PR_SLOWHZ	2
#define PR_FASTHZ	5
/* #define KERNEL */
#define BSD 43
#define TCP_R_COMPAT_42
#ifndef MIN
#define MIN(A,B) ((A) <= (B) ? (A) : (B))
#endif
#ifndef MAX
#define MAX(A,B) ((A) >= (B) ? (A) : (B))
#endif

typedef struct {
    Semaphore	s;
    int		waitCount;
} Tcp_RSem;

struct tcp_rstate {
    u_int	flags;		/* see below */
    struct inpcb *inpcb;
    struct tcp_rcb *tcp_rcb;
    Tcp_RSem	waiting;
    Tcp_RSem	lock;
    XObj	hlpType;	/* for passively opened sessions */
    struct sb	*snd;
    int		rcv_space;	/* amount of space in the receiver's buffer */
    int		rcv_hiwat;	/* size of receiver's buffer */
    int		closed;
};

#define TCP_RST_NBIO		0x0001	/* non-blocking i/o operations */
#define TCP_RST_OOBINLINE		0x0002	/* inline urgent data */
#define TCP_RST_PUSH_ALWAYS	0x0004	/* assume TCP_R-push after every xPush */
#define TCP_RST_RCV_ACK_ALWAYS	0x0008	/* assume infinite receive buffer */

#define sotoinpcb(S) (((struct tcp_rstate *)((S)->state))->inpcb)
#define sototcp_rst(S) ((struct tcp_rstate *)(S)->state)
#define tcp_rcbtoso(tp) (tp->t_inpcb->inp_session)
extern int tracetcp_rp;

/*
 * The arguments to usrreq are:
 *	(*protosw[].pr_usrreq)(up, req, m, nam, opt);
 * where up is a (struct socket *), req is one of these requests,
 * m is a optional mbuf chain containing a message,
 * nam is an optional mbuf chain containing an address,
 * and opt is a pointer to a socketopt structure or nil.
 * The protocol is responsible for disposal of the mbuf chain m,
 * the caller is responsible for any space held by nam and opt.
 * A non-zero return from usrreq gives an
 * UNIX error number which should be passed to higher level software.
 */
#define	PRU_ATTACH		0	/* attach protocol to up */
#define	PRU_DETACH		1	/* detach protocol from up */
#define	PRU_BIND		2	/* bind socket to address */
#define	PRU_LISTEN		3	/* listen for connection */
#define	PRU_CONNECT		4	/* establish connection to peer */
#define	PRU_ACCEPT		5	/* accept connection from peer */
#define	PRU_DISCONNECT		6	/* disconnect from peer */
#define	PRU_SHUTDOWN		7	/* won't send any more data */
#define	PRU_RCVD		8	/* have taken data; more room now */
#define	PRU_SEND		9	/* send this data */
#define	PRU_ABORT		10	/* abort (fast DISCONNECT, DETATCH) */
#define	PRU_CONTROL		11	/* control operations on protocol */
#define	PRU_SENSE		12	/* return status into m */
#define	PRU_RCVOOB		13	/* retrieve out of band data */
#define	PRU_SENDOOB		14	/* send out of band data */
#define	PRU_SOCKADDR		15	/* fetch socket's address */
#define	PRU_PEERADDR		16	/* fetch peer's address */
#define	PRU_CONNECT2		17	/* connect two sockets */
/* begin for protocols internal use */
#define	PRU_FASTTIMO		18	/* 200ms timeout */
#define	PRU_SLOWTIMO		19	/* 500ms timeout */
#define	PRU_PROTORCV		20	/* receive from below */
#define	PRU_PROTOSEND		21	/* send to below */

#if	!defined(KERNEL) || MACH_VMTP
#define PRU_INVOKE		22	/* Invoke a message transaction */
#define PRU_RECVREQ		23	/* Receive a request */
#define PRU_REPLY		24	/* Send a reply */
#define PRU_FORWARD		25	/* Forward a request */
#define PRU_PROBE		26	/* Probe an entity */
#define PRU_GETREPLY		27 	/* Get next reply (multicast) */
#define	PRU_NREQ		27
#else	/* !defined(KERNEL) || MACH_VMTP */
#define PRU_NREQ		21	
#endif /* !defined(KERNEL) || MACH_VMTP */


extern char *prurequests_[];

/*
 * Subset of the inpcb defined in /usr/include/netinet/in_pcb.h
 */
#ifdef XKMACHKERNEL
struct in_addr
	{
		xk_u_int32 s_addr;
	} ;
#endif /* XKMACHKERNEL */

struct inpcb {
	struct	inpcb *inp_next,*inp_prev;
	struct	inpcb *inp_head;
	caddr_t	inp_ppcb;		/* pointer to per-protocol pcb */
	struct in_addr inp_laddr, inp_raddr;
	u_short	inp_lport, inp_rport;
	XObj	inp_session;
};

#define RCV 0
#define SND 1

#ifndef XKMACHKERNEL
#include <sys/errno.h>
#else
#define	EINVAL		22		/* Invalid argument */
#define	EWOULDBLOCK	35		/* Operation would block */
#define	EOPNOTSUPP	45		/* Operation not supported on socket */
#define ECONNRESET	54
#define ENOBUFS		55
#define	EISCONN		56		/* Socket is already connected */
#define ETIMEDOUT	60
#define ECONNREFUSED	61
#endif /* XKMACHKERNEL */

typedef	xk_u_int32	tcp_r_seq;
/*
 * TCP_R header.
 * Per RFC 793, September, 1981.
 */
struct tcp_rhdr {
	u_short	th_sport;		/* source port */
	u_short	th_dport;		/* destination port */
	u_int	th_seq;			/* sequence number */
	u_int	th_ack;			/* acknowledgement number */
	u_int	th_off:4,		/* data offset */
		th_x2:4;		/* (unused) */
	u_char	th_flags;
#define	TH_FIN	0x01
#define	TH_SYN	0x02
#define	TH_RST	0x04
#define	TH_PUSH	0x08
#define	TH_ACK	0x10
#define	TH_URG	0x20
	u_short	th_win;			/* window */
	u_short	th_sum;			/* checksum */
	u_short	th_urp;			/* urgent pointer */
};

#define	TCP_ROPT_EOL	0
#define	TCP_ROPT_NOP	1
#define	TCP_ROPT_MAXSEG	2

/*
 * Default maximum segment size for TCP_R.
 * With an IP MSS of 576, this is 536,
 * but 512 is probably more convenient.
 */
#ifdef	lint
#define	TCP_R_MSS	536
#else
#ifndef IP_MSS
#define	IP_MSS	576
#endif
#define	TCP_R_MSS	MIN(512, IP_MSS - sizeof (struct tcp_riphdr))
#endif

/*
 * User-settable options (used with setsockopt).
 */
#define	TCP_R_NODELAY	0x01	/* don't delay send to coalesce packets */
#define	TCP_R_MAXSEG	0x02	/* set maximum segment size */


/*
 * X-Kernel defines
 */

/*====================  #8 10 lines added 070390 cliff ===================*/

#define TCP_R_IPSEC		/* Compiles with code necessary for IPSEC */
#define IPSEC		        /* Compiles with code necessary for RFC1825 */

typedef struct pstate  {
    IPhost	myHost;
    Map     	activeMap;
    Map      	passiveMap;
    VOID	*portstate;
    struct tcp_rstat	*tcp_rstat;	/* tcp_r statistics */
    struct inpcb	*tcb;		/* head of queue of active tcp_rcb's */
#ifdef TCP_R_IPSEC
    int 	useIpsec;
#endif	
}  PSTATE;


typedef TCP_Rport  PassiveId;

typedef struct {
      unsigned short localport;
      unsigned short remoteport;
      IPhost         remoteaddr;
} ActiveId;

typedef struct {
	IPpseudoHdr	*h;
	Msg		*m;
} hdrStore_t;

#define FROM_ENABLE	-1
#define TCP_RRCVWIN	(TCP_R_BUFFER_SPACE)

extern char *tcp_rstates[];

#if defined(__STDC__) || defined(__GNUC__)

/*
 * in_hacks.c
 */
int	in_pcballoc_( XObj );
void	in_pcbdisconnect_( struct inpcb * );
void 	in_pcbdetach_( struct inpcb * );
bool	in_broadcast_( struct in_addr );
u_short	in_cksum( Msg *, u_short *, int len );

/*
 * tcp_r_subr.c
 */
struct tcp_riphdr *	tcp_r_template( struct tcp_rcb * );
struct tcp_rcb *	tcp_r_newtcp_rcb( struct inpcb * );
struct tcp_rcb *	tcp_r_drop( struct tcp_rcb *, int errnum );
struct tcp_rcb * 	tcp_r_destroy( struct tcp_rcb * );
void	tcp_r_notify( struct inpcb * );
void	tcp_r_quench( struct inpcb * );
void	tcp_r_respond( struct tcp_rcb *, struct tcp_rhdr *, IPpseudoHdr *,
			     tcp_r_seq ack, tcp_r_seq seq, int flags,
		             XObj tcp_rProtl );
 
/*
 * tcp_r_hdr.c
 */
void 	tcp_rHdrStore(void *, char *, long, void *);
void 	tcp_rOptionsStore(void *, char *, long, void *);
long	tcp_rOptionsLoad(void *, char *, long, void *);
long 	tcp_rHdrLoad(void *, char *, long, void *);

/*
 * tcp_r_timer.c
 */
void	tcp_r_fasttimo( Event, void * );
void	tcp_r_slowtimo( Event, void * );
void	tcp_r_canceltimers( struct tcp_rcb * );
struct tcp_rcb *	tcp_r_timers( struct tcp_rcb *, int );

/*
 * tcp_r_x.c
 */
void	delete_tcp_rstate( struct tcp_rstate * );
struct tcp_rstate	*	new_tcp_rstate( void );
void	sorwakeup_( XObj );
void	sowwakeup_( XObj );
void	soabort_( XObj );
void	socantsendmore_( XObj );
void	socantrcvmore_( XObj );
void	sohasoutofband_( XObj, u_int );
void	soisdisconnected_( XObj );
void	soisdisconnecting_( XObj );
void	soisconnected_( XObj );
void	soisconnecting_( XObj );
int	soreserve_( XObj, int, int );
#ifdef TCP_R_IPSEC
XObj	sonewconn_( XObj, XObj, XObj, XObj, IPhost *, IPhost *, int, int );
#else
XObj	sonewconn_( XObj, XObj, XObj, IPhost *, IPhost *, int, int );
#endif
void	tcp_rSemWait( Tcp_RSem * );
void	tcp_rSemSignal( Tcp_RSem * );
void	tcp_rSemInit( Tcp_RSem *, int );
void	tcp_rVAll( Tcp_RSem * );

/* 
 * tcp_r_output.c
 */
int	tcp_r_output( struct tcp_rcb * );
void	tcp_r_setpersist( struct tcp_rcb * );

/* 
 * tcp_r_usrreq.c
 */
int	tcp_r_attach( XObj );
int	tcp_r_usrreq( XObj, int, Msg *, Msg * );
struct tcp_rcb *tcp_r_disconnect(struct tcp_rcb * );
struct tcp_rcb *tcp_r_usrclosed( struct tcp_rcb * );

/* 
 * tcp_r_input.c
 */
void	print_reass_( struct tcp_rcb *, char * );
xkern_return_t	tcp_r_input( XObj, XObj, Msg * );
int	tcp_r_mss( struct tcp_rcb * );
xkern_return_t	tcp_rPop( XObj, XObj, Msg *, VOID * );
int	tcp_r_reass( struct tcp_rcb *, struct tcp_rhdr *, XObj, Msg *, Msg * );

/* 
 * tcp_r_debug.c
 */
void	tcp_r_trace(int, int, struct tcp_rcb *, struct tcp_riphdr *, int );
char *	tcp_rFlagStr( int );

#else

/*
 * in_hacks.c
 */
int	in_pcballoc_();
void	in_pcbdisconnect_();
void 	in_pcbdetach_();
bool	in_broadcast_();
u_short	in_cksum();

/*
 * tcp_r_subr.c
 */
struct tcp_riphdr *	tcp_r_template();
struct tcp_rcb *	tcp_r_newtcp_rcb();
struct tcp_rcb *	tcp_r_drop();
struct tcp_rcb * 	tcp_r_destroy();
void	tcp_r_notify();
void	tcp_r_quench();
void	tcp_r_respond();
 
/*
 * tcp_r_hdr.c
 */
void 	tcp_rHdrStore();
void 	tcp_rOptionsStore();
long	tcp_rOptionsLoad();
long 	tcp_rHdrLoad();

/*
 * tcp_r_timer.c
 */
void	tcp_r_fasttimo();
void	tcp_r_slowtimo();
void	tcp_r_canceltimers();
struct tcp_rcb *	tcp_r_timers();

/*
 * tcp_r_x.c
 */
void	delete_tcp_rstate();
struct tcp_rstate	*	new_tcp_rstate();
void	sorwakeup_();
void	sowwakeup_();
void	soabort_();
void	socantsendmore_();
void	socantrcvmore_();
void	sohasoutofband_();
void	soisdisconnected_();
void	soisdisconnecting_();
void	soisconnected_();
void	soisconnecting_();
int	soreserve_();
XObj	sonewconn_();
void	tcp_rSemWait();
void	tcp_rSemSignal();
void	tcp_rSemInit();
void	tcp_rVAll();

/* 
 * tcp_r_output.c
 */
int	tcp_r_output();
void	tcp_r_setpersist();

/* 
 * tcp_r_usrreq.c
 */
int	tcp_r_attach();
int	tcp_r_usrreq();
struct tcp_rcb *tcp_r_disconnect();
struct tcp_rcb *tcp_r_usrclosed();

/* 
 * tcp_r_input.c
 */
void	print_reass_();
xkern_return_t	tcp_r_input();
int	tcp_r_mss();
xkern_return_t	tcp_rPop();
int	tcp_r_reass();

/* 
 * tcp_r_debug.c
 */
void	tcp_r_trace();
char *	tcp_rFlagStr();

#endif /* __STDC__ */

#include "insque.h"
