/*
 * tcp_usrreq.c
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_usrreq.c	7.6 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:39:21 $
 */

#include "xkernel.h"
#include "tcp_internal.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"
#include "tcp_debug.h"

/*
 * TCP_S protocol interface to socket abstraction.
 */
extern	char *tcp_sstates[];
int	tcp_ssenderrors;

/*
 * Process a TCP_S user request for TCP_S tb.  If this is a send request
 * then m is the mbuf chain of send data.  If this is a timer expiration
 * (called from the software clock routine), then timertype tells which timer.
 */
/*ARGSUSED*/
int
tcp_s_usrreq(so, req, m, nam)
	XObj so;
	int req;
	Msg *m, *nam;
{
	register struct inpcb *inp;
	register struct tcp_scb *tp = 0;
	register struct tcp_sstate *tcp_sst = (struct tcp_sstate*)so->state;
	/* int s; */
	int error = 0;
	int ostate;

	xTrace0(tcp_sp, 7, "tcp_s_usrreq");
	/* 
	 * The original code protected this routine with splnet().  If
	 * this thread blocks, there could be a problem.
	 */
 	/* s = splnet(); */
	inp = sotoinpcb(so);
	/*
	 * When a TCP_S is attached to a socket, then there will be
	 * a (struct inpcb) pointed at by the socket, and this
	 * structure will point at a subsidary (struct tcp_scb).
	 */
	if (inp == 0 && req != PRU_ATTACH) {
		/* splx(s); */
		return (EINVAL);		/* XXX */
	}
	if (inp) {
		tp = intotcp_scb(inp);
		/* WHAT IF TP IS 0? */
#ifdef KPROF
		tcp_s_acounts[tp->t_state][req]++;
#endif
		ostate = tp->t_state;
	} else
		ostate = 0;
	switch (req) {

	/*
	 * TCP_S attaches to socket via PRU_ATTACH, reserving space,
	 * and an internet control block.
	 */
	case PRU_ATTACH:
		if (inp) {
			error = EISCONN;
			break;
		}
		error = tcp_s_attach(so);
		if (error)
			break;
		tp = sototcp_scb(so);
		break;

	/*
	 * PRU_DETACH detaches the TCP_S protocol from the socket.
	 * If the protocol state is non-embryonic, then can't
	 * do this directly: have to initiate a PRU_DISCONNECT,
	 * which may finish later; embryonic TCB's can just
	 * be discarded here.
	 */
	case PRU_DETACH:
		if (tp->t_state > TCP_SS_LISTEN)
			tp = tcp_s_disconnect(tp);
		else
			tp = tcp_s_destroy(tp);
		break;

	/*
	 * Initiate connection to peer.
	 * Create a template for use in transmissions on this connection.
	 * Enter SYN_SENT state, and mark socket as connecting.
	 * Start keep-alive timer, and seed output sequence space.
	 * Send initial segment on connection.
	 */
	case PRU_CONNECT:
	    {
		struct	tcp_sstat *tcp_sstat;

		tcp_sstat = ((PSTATE *)xMyProtl(so)->state)->tcp_sstat;
		xTrace0(tcp_sp, 7, "usrreq: PRU_CONNECT");
		tp->t_template = tcp_s_template(tp);
		if (tp->t_template == 0) {
			in_pcbdisconnect(inp);
			error = ENOBUFS;
			break;
		}
		soisconnecting(so);
		tcp_sstat->tcp_ss_connattempt++;
		tp->t_state = TCP_SS_SYN_SENT;
		tp->t_timer[TCP_ST_KEEP] = TCP_STV_KEEP;
		tp->iss = tcp_s_iss; tcp_s_iss += TCP_S_ISSINCR/2;
		tcp_s_sendseqinit(tp);
		error = tcp_s_output(tp);
		break;
	    }

	/*
	 * Create a TCP_S connection between two sockets.
	 */
	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	/*
	 * Initiate disconnect from peer.
	 * If connection never passed embryonic stage, just drop;
	 * else if don't need to let data drain, then can just drop anyways,
	 * else have to begin TCP_S shutdown process: mark socket disconnecting,
	 * drain unread data, state switch to reflect user close, and
	 * send segment (e.g. FIN) to peer.  Socket will be really disconnected
	 * when peer sends FIN and acks ours.
	 *
	 * SHOULD IMPLEMENT LATER PRU_CONNECT VIA REALLOC TCP_SCB.
	 */
	case PRU_DISCONNECT:
		tp = tcp_s_disconnect(tp);
		break;

	/*
	 * Accept a connection.  Essentially all the work is
	 * done at higher levels; just return the address
	 * of the peer, storing through addr.
	 */
	case PRU_ACCEPT: {
#if 0
		struct sockaddr_in *sin = mtod(nam, struct sockaddr_in *);

		nam->m_len = sizeof (struct sockaddr_in);
		sin->sin_family = AF_INET;
		sin->sin_port = inp->inp_fport;
		sin->sin_addr = inp->inp_faddr;
#endif
		break;
		}

	/*
	 * After a receive, possibly send window update to peer.
	 */
	case PRU_RCVD:
		(void) tcp_s_output(tp);
		break;

	/*
	 * Do a send by putting data in output queue and updating urgent
	 * marker if URG set.  Possibly send more data.
	 */
	case PRU_RCVOOB:
#if 0
		if ((so->so_oobmark == 0 &&
		    (so->so_state & SS_RCVATMARK) == 0) ||
#ifdef SO_OOBINLINE
		    so->so_options & SO_OOBINLINE ||
#endif
		    tp->t_oobflags & TCP_SOOB_HADDATA) {
			error = EINVAL;
			break;
		}
#endif
		if ((tp->t_oobflags & TCP_SOOB_HAVEDATA) == 0) {
			error = EWOULDBLOCK;
			break;
		}
		msgDestroy(m);
		msgConstructBuffer(m, &tp->t_iobc, 1);
		tp->t_oobflags ^= (TCP_SOOB_HAVEDATA | TCP_SOOB_HADDATA);
		break;

	case PRU_SENDOOB:
		/*
		 * According to RFC961 (Assigned Protocols),
		 * the urgent pointer points to the last octet
		 * of urgent data.  We continue, however,
		 * to consider it to indicate the first octet
		 * of data past the urgent section.
		 * Otherwise, snd_up should be one lower.
		 */
		sbappend(tcp_sst->snd, m);
		tp->snd_up = tp->snd_una + sblength(tcp_sst->snd);
		tp->t_force = 1;
		error = tcp_s_output(tp);
		tp->t_force = 0;
		break;

	/*
	 * TCP_S slow timer went off; going through this
	 * routine for tracing's sake.
	 */
	case PRU_SLOWTIMO:
		tp = tcp_s_timers(tp, (int)nam);
		req |= (int)nam << 8;		/* for debug's sake */
		break;

	default:
		Kabort("tcp_s_usrreq");
	}
	xIfTrace(tcp_sp, 3) {
		tcp_s_trace(TA_USER, ostate, tp, (struct tcp_siphdr *)0, req);
	}
	/* splx(s); */
	return (error);
}

int	tcp_s_sendspace = TCP_SRCVWIN;
int	tcp_s_recvspace = TCP_SRCVWIN;
/*
 * Attach TCP_S protocol to socket, allocating
 * internet protocol control block, tcp_s control block,
 * bufer space, and entering LISTEN state if to accept connections.
 */
int
tcp_s_attach(so)
	XObj so;
{
	register struct tcp_scb *tp;
	struct inpcb *inp;
	int error;

	error = soreserve(so, tcp_s_sendspace, tcp_s_recvspace);
	if (error)
		return (error);
	error = in_pcballoc(so);
	if (error)
		return (error);
	inp = sotoinpcb(so);
	tp = tcp_s_newtcp_scb(inp);
	tp->t_state = TCP_SS_CLOSED;
	return (0);
}

/*
 * Initiate (or continue) disconnect.
 * If embryonic state, just send reset (once).
 * If in ``let data drain'' option and linger null, just drop.
 * Otherwise (hard), mark socket disconnecting and drop
 * current input data; switch states based on user close, and
 * send segment to peer (with FIN).
 */
struct tcp_scb *
tcp_s_disconnect(tp)
	register struct tcp_scb *tp;
{
	XObj so = tp->t_inpcb->inp_session;
	xTrace1(tcp_sp, 3, "tcp_s_disconnect: tp %X", tp);
	if (tp->t_state < TCP_SS_ESTABLISHED)
		tp = tcp_s_destroy(tp);
	else {
		soisdisconnecting(so);
		tp = tcp_s_usrclosed(tp);
		if (tp)
			(void) tcp_s_output(tp);
	}
	return (tp);
}

/*
 * User issued close, and wish to trail through shutdown states:
 * if never received SYN, just forget it.  If got a SYN from peer,
 * but haven't sent FIN, then go to FIN_WAIT_1 state to send peer a FIN.
 * If already got a FIN from peer, then almost done; go to LAST_ACK
 * state.  In all other cases, have already sent FIN to peer (e.g.
 * after PRU_SHUTDOWN), and just have to play tedious game waiting
 * for peer to send FIN or not respond to keep-alives, etc.
 * We can let the user exit from the close as soon as the FIN is acked.
 */
struct tcp_scb *
tcp_s_usrclosed(tp)
	register struct tcp_scb *tp;
{

	switch (tp->t_state) {

	case TCP_SS_CLOSED:
	case TCP_SS_LISTEN:
	case TCP_SS_SYN_SENT:
		tp->t_state = TCP_SS_CLOSED;
		tp = tcp_s_destroy(tp);
		break;

	case TCP_SS_SYN_RECEIVED:
	case TCP_SS_ESTABLISHED:
		tp->t_state = TCP_SS_FIN_WAIT_1;
		break;

	case TCP_SS_CLOSE_WAIT:
		tp->t_state = TCP_SS_LAST_ACK;
		break;
	}
	if (tp && tp->t_state >= TCP_SS_FIN_WAIT_2)
		soisdisconnected(tp->t_inpcb->inp_session);
	return (tp);
}
