/*     
 * $RCSfile: udptest.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Revision: 1.1 $
 * $Date: 1996/01/24 01:26:07 $
 */



/* this should be in part.h or something */

#define PART_TYPE_SECURE 4

/*
 * Ping-pong test of UDP
 */

#include "part.h"
#include "udp.h"
#include "saidmgr.h"
#include "ipsec.h"

/*
 * These definitions describe the lower protocol
 */
#define HOST_TYPE IPhost
#define INIT_FUNC udptest_init
#define TRACE_VAR traceudptestp
#define PROT_STRING "udp"

extern XObj	SAIDPROT;
extern XObj	MYPROT;

/* 
 * If a host is booted without client/server parameters and matches
 * one of these addresses, it will come up in the appropriate role.
 */
static HOST_TYPE ServerAddr = { 0, 0, 0, 0 };
static HOST_TYPE ClientAddr = { 0, 0, 0, 0 };
long	clientPort = ANY_PORT;

static long	serverPort = 2001;

#define TRIPS 100
#define TIMES 1
#define DELAY 3
/*
 * Define to do timing calculations
 */
#define TIME

static XObj
spiOpen(buf, mode)
  char *buf; XIpsecMode mode;
{
  XAlgorithm algo;

    Part pi[SPI_PART+1];

    partInit(pi, SPI_PART+1);
    algo = xControl(SAIDPROT,SAIDMGR_RESOLVE_NAME,buf,strlen(buf));
    partInit(pi,SPI_PART+1);
    partPush(pi[REMOTE_PART], &ServerAddr, sizeof(HOST_TYPE));
    partPush(pi[LOCAL_PART], &ClientAddr, sizeof(HOST_TYPE));
    partPush(pi[SPI_PART], &mode, sizeof(XIpsecMode));
    partPush(pi[SPI_PART], &algo, sizeof(XAlgorithm));
    
    printf("Start open saidmgr\n");
    return xOpen(MYPROT, MYPROT, SAIDPROT,pi);
  }

/* 
 * It may be convenient to send between different UDP ports on the
 * same host, so we'll have custom participant assignment
 */
#define CUSTOM_ASSIGN
#define CUSTOM_OPENDONE

extern char *testalg;
extern char *testipsecmode;

static void
clientSetPart( p )
    Part	*p;
{
    static XSaid said=0, said2=0;
    static int   ptysec = - PART_TYPE_SECURE;
    XObj lls;
    XAlgorithm algo;
    XIpsecMode mode = IPSEC_ESP_ENCAPS;
/*    XIpsecMode mode = IPSEC_AH; */
/*    XIpsecMode mode = IPSEC_ESP_TUNNEL; */
    Part pi[3];
    char buf[100];

/*    strcpy(buf,"crypt"); */
    if (!testalg)    strcpy(buf,"md5"); 
    else strcpy(buf, testalg);
    printf("client using algorithm %s\n", buf);
    if (testipsecmode) {
      if (!strcmp(testipsecmode, "esp_encaps")) mode = IPSEC_ESP_ENCAPS;
      else if (!strcmp(testipsecmode, "esp_tunnel")) mode = IPSEC_ESP_TUNNEL;
      else if (!strcmp(testipsecmode, "ah")) mode = IPSEC_AH;
      printf("client using mode %s (%d)\n", testipsecmode, mode);
    }
    else printf("client using mode (%d)\n", mode);
    lls = spiOpen(buf, mode);
    if (lls != ERR_XOBJ) {
      printf("Test client: Open finished %x\n",lls);
      printf("Got local SAID %u\n",said=xControl(lls,SAIDMGR_GETLSAID,NULL,0)); 
      if ((lls = spiOpen("md5", IPSEC_AH)) == ERR_XOBJ) {
	printf("Test client: Second open failed %x\n",lls);
	return;
      }
      printf("Test client: Second open finished %x\n",lls);
      printf("Got local SAID %u\n",said2=xControl(lls,SAIDMGR_GETLSAID,NULL,0)); 

      partInit(p, SPI_PART+1);
      partPush(p[REMOTE_PART], &ServerAddr, sizeof(IPhost));
      partPush(p[REMOTE_PART], &serverPort, sizeof(long));
      {
	/* 
	 * NOTE -- if you use two participants, make sure the second argument to 
	 * the above call to partInit is 3
	 */
	partPush(p[LOCAL_PART], &ClientAddr, 0);
	partPush(p[LOCAL_PART], &clientPort, sizeof(long));
      }
/*      partPush(p[SPI_PART], &said2, sizeof(XSaid)); /* ah */
      partPush(p[SPI_PART], &said, sizeof(XSaid));
      partPush(p[SPI_PART], (void *)ptysec, 0);
    }
    else printf("Client could not open SAIDMGR\n");
}

static void
serverSetPart( p )
    Part 	*p;
{
    partInit(p, 1);
    partPush(*p, ANY_HOST, 0);
    partPush(*p, &serverPort, sizeof(long));
}

static int lens[] = { 
/*  1, 1000, 2000, 4000, 8000, 16000 */
  2000, 4000, 8000, 16000
};

static xkern_return_t
customOpenDone( self, lls, llp, hlpType )
    XObj self, lls, llp, hlpType;
{
  Part parts[SPI_PART+1];
  char pbuf[3*sizeof(parts)];  /* ??? */

  printf("server custom opendone ");
  partInit(parts, SPI_PART+1);
  if (  xControl(lls, GETPARTICIPANTS, pbuf, sizeof(pbuf)) > 1) {
    partInternalize(parts, pbuf);
    while (partStackTopByteLen(parts[SPI_PART]) > 0)
      printf("opendone spi %d ", *(int *)partPop(parts[SPI_PART]));
  }
  printf("\n");
  xDuplicate(lls);
}

#define SAVE_SERVER_SESSN

#include "common_test.c"


static void
testInit()
{
}

