/*********************************************************************

 Bellcore Proprietary - For Bellcore and Authorized Clients only

 This software is proprietary information that shall be distributed or
 routed only within Bell Communications Research (Bellcore) and its
 authorized clients, except with written permission of Bellcore.

 Copyright 1991 Bell Communications Research, Morristown, NJ
 All rights reserved.
 Any unauthorized copying or use is prohibited,
 without the prior written permission of
 Bell Communications Research (Bellcore).
 This copyright notice must accompany this software.
 *******************************************************************/


/*********************************************************************
 * sarCtlr.c
 * Device driver code for SAR interface board to Sunshine switch
 * using TurboChannel on DEC 5000 Workstation
 *
 * 5/18/94: Peter Druschel, modified for Scout on DEC Alpha
 *
 * 5/3/92: Peter Druschel, modified for x-kernel in Mach3.0.
 *
 * Created: 7/1/91  Michael DeAddio from Ultrix manual for writing
 * device drivers.
 *
 * $RCSfile: sarCtlr.c,v $
 *
 * HISTORY
 * $Log: sarCtlr.c,v $
 * Revision 1.2  1994/06/04 05:05:28  druschel
 * Lots of fixes and cleanup
 *
 * Revision 1.1  1994/05/22  22:25:39  druschel
 * Initial revision
 *
 *
 *******************************************************************/

#define SIM_RCV_HACK
/*#define SEQNUM*/
/*#define ZERO_RCV_BUFFERS*/
/*#define IPHDR_CHECK*/
/*#define DPM_TEST*/
/*#define TEST_HDR_INTR*/

/* include files */
#include "scout_device.h"
#include "scout_sync_arch.h"
#include "scout_thread.h"
#include "scout_util.h"
#include "xkernel.h"
#include "eth_i.h"
#include "sarCtlr.h"

/* prototypes */
u_int kvtophys(char* va);
void dpmWriteInt(volatile int *loc, int data);
void dpmWritePtr(volatile u_int *loc, u_int ptr);

xkern_return_t sar_init(XObj self);
static xmsg_handle_t push(XObj s, Msg *msg);
static xkern_return_t open_enable(XObj self, XObj hlp, XObj hlptype, Part *part);
static int prot_control(XObj self, int op, char *buf, int len);

int sarprobe(SarSoftc *soft);
int sarCtlrInit(XObj self);
void initRcvBQueue(SarSoftc *soft);
void sarRcv(Event *ev, XObj self);
void sarFreeRcvBuf(void* virtAddr, int len);
void sar_intr(XObj self);
void sarReset (XObj self);
int semaLock(SarReg *sreg, volatile int *sema);
void semaUnlock(SarReg *sreg, volatile int *sema);
static void sarXmitUnwire(char *chPtr, long chLen, VOID *arg);
static bool sarXmitCollect(char *virtAddr, long	len, VOID *arg);
int sarCtlrXmit(Msg *downMsg, int vci, int *seqNum);
int sarGetStatistics(int *xmitStats, int* rcvStats);

/* driver state */
SarSoftc sar_softc;


/*
 * init tracesar
 * sarprobe and sarattach run before initTraceLevel() gets called!
 */
int tracesarp = TR_ERRORS;
#ifdef IPHDR_CHECK
int ip_hdr_check = 0;
#endif

/* a temporary hack */
void lance_interrupt(XObj self) {}

/* 
 * physical address convertion
 */
inline u_int
  kvtophys(char* va)
{
  return (u_int)K0SEG_TO_P(va);
}

#define MACHINE_PGBYTES 16384



/*
 * DPM write functions
 * these functions encapsulate a workaround for a suspected hw bug
 */

void
  dpmWriteInt(loc, data)
volatile int *loc;
int data;
{
  do {
    *loc = data;
    scout_flush_write_to_memory();
    /*if (*loc != data) printf("DPM bug: wrote int %x, read %x\n", data, *loc);*/
  } while (*loc != data);
}

void
  dpmWritePtr(loc, ptr)
volatile u_int *loc;
u_int ptr;
{
  do {
    *loc = ptr;
    scout_flush_write_to_memory();
    /*if (*loc != ptr) printf("DPM bug: wrote ptr %x, read %x\n", ptr, *loc);*/
  } while (*loc != ptr);
}


/*
 *
 * x-kernel/Scout boilerplate
 * 
 */


xkern_return_t
sar_init(XObj self)
{
    SarSoftc *soft = (SarSoftc*) self->state;
    SarReg *regs;
    const Scout_Device_Info *di;

    xTrace1(sarp, TR_FUNCTIONAL_TRACE, "sar_init(self=%p)", self);
    if (!soft)  {
	soft = &sar_softc;
	bzero(soft, sizeof(*soft));
	self->state = soft;

	if (!self->instName) {
	    self->instName = "SAR-A";
	} /* if */

	/* find device: */
	di = scout_device_info(self->instName);
	if (!di) {
	    xTrace1(sarp, TR_DETAILED,
		    "sar_init: unknown device %s", self->instName);
	    xError("sar_init: unknown device.");
	    return XK_FAILURE;
	} /* if */

	soft->sc_regs = regs = (SarReg*)(di->base_addr);
	soft->my_ipl = di->ipl;
	strcpy(soft->sc_devname, di->name);        /* device name */	
	soft->sc_unit = 0;
				
	/* init the fake eth hdr for incoming packets */
	soft->ethHdr.dst.high = regs->ethAddr.high;
	soft->ethHdr.dst.mid = regs->ethAddr.mid;
	soft->ethHdr.dst.low = regs->ethAddr.low;
	if (regs->ethAddr.low == htons(0xab3e)) {
	  /* set the src to joplin's eth address */
	  soft->ethHdr.src.high = htons(0x0800);
	  soft->ethHdr.src.mid = htons(0x2b3f);
	  soft->ethHdr.src.low = htons(0x53bc);
	}
	else {
	  ZERO_ETH_AD(soft->ethHdr.src);
	}
	/* soft->ethHdr.type = htons(0x3900);  IP */
	soft->ethHdr.type = htons(0x3000); /* ethtest */


	/* enable SAR interrupts and set interrupt arg: */
	scout_device_set_interrupt(di, 1, sar_intr, (void*)self);

	if (sarprobe(soft) && sarCtlrInit(self)) {
	    self->openenable = open_enable;
	    self->push	     = push;
	    self->control    = prot_control;

	    xTrace0(sarp, TR_FUNCTIONAL_TRACE,
		     "sar_init: returning successfully");
	    return XK_SUCCESS;
	} /* if */
    } else {
	xError("sar_init: driver initialized already.");
    } /* if */
    return XK_FAILURE;
} /* sar_init */


static int seqNum = 0;

static xmsg_handle_t
push(XObj s, Msg *msg)
{
    XObj self = xMyProtl(s);
    SarSoftc *soft = (SarSoftc*) self->state;
    ETHhdr *hdr = (ETHhdr*) msgGetAttr(msg, 0);
    /* using VCI 1 */
    int vci = 1;
    
    if (msgLen(msg) > SARMTU) {
      xTrace1(sarp, TR_ERRORS, "sarPush::message is too big, len=%d", 
	       msgLen(msg));
      return XMSG_NULL_HANDLE;
    }

    sarCtlrXmit(msg, vci, &seqNum);

    return XMSG_NULL_HANDLE;
} /* push */


static xkern_return_t
open_enable(XObj self, XObj hlp, XObj hlptype, Part *part)
{
    xTraceP0(self, TR_FUNCTIONAL_TRACE, "open_enable");
    /*
     * Shepherd threads invoke xDemux() on our *protocol* (normally,
     * xDemux() should be invoked on sessions only).  xDemux() in
     * turn uses an XObj's "up" field to determine the protocol
     * to which to demux.  So all we do here is set this "up" pointer:
     */
    xSetUp(self, hlp);
    return XK_SUCCESS;
} /* open_enable */


static int
prot_control(XObj self, int op, char *buf, int len)
{
  SarSoftc *soft = (SarSoftc*) self->state;

    xTraceP1(self, TR_FULL_TRACE, "prot_control(op=%#x)", op);

    switch (op) {
      case GETMYHOST:
        if (len >= sizeof(ETHhost)) {
	    SarReg *sreg;

	    if (!soft) {
		xError("sar: prot_control: protocol not initialized");
		return -1;
	    } /* if */

	    sreg = soft->sc_regs;
	    bcopy( (char*)&(sreg->ethAddr), buf, sizeof(ETHhost) );

	    xTraceP1(self, TR_DETAILED, "prot_control: returning addr %s",
		     ethHostStr(&(sreg->ethAddr)));
	    return sizeof(ETHhost);
	} /* if */
	break;
      default:
	break;
    } /* switch */
    return -1;
} /* prot_control */


/*
 *
 * end of x-kernel/Scout boilerplate
 * 
 */


int sarprobe(soft) 
  SarSoftc *soft;
{
  int i, ntries, val;
  unsigned j;
  volatile int *ptr, *mem = NULL;
  u_char *foo;
  SarReg *sreg = soft->sc_regs;

  xTrace1(sarp, TR_MAJOR_EVENTS, "Inside sarprobe, regs %lx.", (u_long)sreg);

  ptr = (volatile int *) &sreg->xmitCSR;

#ifdef DPM_TEST
  /* test of dual-ported memory */
  for (i = 0 ; i < DPMEM_SIZE; i++) {
    int sav = ptr[i]; 

    for (j = 1; j != 0; j <<= 1) {
      ptr[i]  = j ;
      scout_flush_write_to_memory();
      if ((val = ptr[i]) != j) {
	xTrace3(sarp, TR_ERRORS, 
	"sarprobe: Dual-ported memory location %x returns %x instead of %x.",
		i, val, j) ;
	mem = &ptr[i] ;
      }
    }

    ptr[i] = sav;
  } 
#endif

  if (mem != NULL) {
    printf ("sarprobe: Dual-ported memory failed R/W test, location %lx.\n", 
	    (u_long)mem) ;
  }
  else {
    printf ("\nsarprobe: Dual-ported memory PASSED R/W test.\n\n") ;
  }

  /* 
   * reset both sar processors
   */
  semaLock(sreg, &sreg->xmitSema);
  dpmWriteInt(&sreg->xmitCSR, SAR_RESET); /* tell write processor to reset */
  semaUnlock(sreg, &sreg->xmitSema);
  
  semaLock(sreg, &sreg->rcvSema);
  dpmWriteInt(&sreg->rcvCSR, SAR_RESET); /* tell read processor to reset */
  semaUnlock(sreg, &sreg->rcvSema);

  /* wait for SAR_RDY */
  ntries = 0 ;
  while ( !(sreg->xmitCSR & SAR_RDY) ) {
    /* check to see if we've exceeded max number of retries*/
    if (ntries++ >= MAX_RETRY) {
      printf("sarprobe: SAR Xmitter not responding.\n");
      break;
    }
    scout_udelay(5);
  }
  xAssert( (sreg->xmitCSR & ~SAR_CSR_MASK) == 0);

  ntries = 0 ;
  while ( !(sreg->rcvCSR & SAR_RDY) ) {
    /* check to see if we've exceeded max number of retries*/
    if (ntries++ >= MAX_RETRY) {
	printf("sarprobe: SAR Receiver not responding.\n");
	break;
    }
    scout_udelay(5);
  }
  xAssert( (sreg->rcvCSR & ~SAR_CSR_MASK) == 0);
  
  xTrace0(sarp, TR_EVENTS, "Exiting sarprobe.") ;
  return 1;
}




/*
 * sarCtlrInit()
 * 
 * Returns:
 * 1 if sucess, 0 otherwise
 */ 
int 
sarCtlrInit(XObj self)
{
  SarSoftc *soft = (SarSoftc*)self->state;
  SarReg *sreg ;
  int i, j;

  xTrace0(sarp, TR_EVENTS, "sarCtlrInit.");

  /* get pointer to sar registers and memory */
  /* init pointer to SAR regs */
  sreg = soft->sc_regs ;

  /* create threads to handle incoming messages */
  for (i = 0; i < 1 /*MAX_XK_THREADS*/; i++) {
    evDetach(evSchedule(sarRcv, self, 0));
  }

  /* init the transmitted msg array */
  for (i = 0; i < XMIT_QUEUE_SIZE; i++) 
    msgConstructEmpty(&soft->xmitMsg[i]);

  /* allocate a page for dumping cells via DMA */
  /* kalloc allocates wired memory */
  sreg->rcvDumpAddr = kvtophys(scout_malloc_aligned(4096,12));

  initRcvBQueue(soft);

  return 1;
}


/*
 * initRcvBQueue()
 * initalize the receiver's free buffer queue
 */
void
initRcvBQueue(soft)
     SarSoftc *soft;
{
  SarReg *sreg ;
  char *virtAddr;
  u_int physAddr;
  int i, j;

  sreg = soft->sc_regs;

#ifndef SIM_RCV_HACK
  /* for now, assume SARMTU <= pageSize */
  xAssert(SARMTU <= MACHINE_PGBYTES);
#endif

  /* allocate wired pages used for read buffers */
  virtAddr = scout_malloc_aligned(FREEB_QUEUE_SIZE * MACHINE_PGBYTES, 13);
  
  /* allocate a msg page pool for read buffers  */
  for (i = 0; i < FREEB_QUEUE_SIZE; i++, virtAddr += MACHINE_PGBYTES) {

    /* allocate a msg buffer */
    soft->rcvFreeBuf[i].virtAddr = virtAddr;
    
    /* write physical address of msg buf to free buffer queue */
    xTrace1(sarp, TR_FULL_TRACE, "sarRcvBQueue:msg buf at %lx", (u_long)virtAddr);

    /* translate to physical address */
    physAddr = soft->rcvFreeBuf[i].physAddr = kvtophys(virtAddr);
    xAssert(physAddr);

    /*sreg->freeBQueue[i].physAddr = physAddr;*/
    dpmWritePtr(&(sreg->freeBQueue[i].physAddr), physAddr);

    /*sreg->freeBQueue[i].bufIndex = i;*/
    dpmWriteInt(&(sreg->freeBQueue[i].bufIndex), i);
    
    /*sreg->freeBQueue[i].size = MACHINE_PGBYTES;*/
    dpmWriteInt(&(sreg->freeBQueue[i].size), MACHINE_PGBYTES);
  }

  /*sreg->freeBQHead = FREEB_QUEUE_SIZE - 1;*/
  dpmWriteInt(&(sreg->freeBQHead), FREEB_QUEUE_SIZE - 1);
  /*sreg->freeBQTail = 0;*/
  dpmWriteInt(&(sreg->freeBQTail), 0);

  semaLock(sreg, &sreg->rcvSema);
  /*sreg->rcvCSR &= ~SAR_NEEDB;*/
  dpmWriteInt(&(sreg->rcvCSR), sreg->rcvCSR & ~SAR_NEEDB);
  semaUnlock(sreg, &sreg->rcvSema);
}


/* 
 * sarRcv()
 * This routine is executed by receiver threads, which are signalled 
 * by the SAR interrupt handler 
 * when data is available.
 *
 * Returns:
 * nothing
 */
void 
sarRcv(Event *ev, XObj self)
{
  SarSoftc *soft = (SarSoftc*)self->state;
  SarReg *sreg;
  Scout_IPL old_ipl;
  int tail, j;

  xTrace0(sarp, TR_EVENTS, "in sarRcv.");

  /* init pointer to SAR regs */
  sreg = soft->sc_regs ;

  soft->rcvThread = scout_thread_self();

  while(1) {
    xTrace0(sarp, TR_EVENTS, "sarRcv thread blocking.") ;
    
    /* wait until more datagrams arrive */
    old_ipl = scout_swap_ipl(soft->my_ipl);
    if (sreg->rcvQTail == sreg->rcvQHead) {
      /* go to sleep */
      scout_swap_ipl(old_ipl);
      scout_thread_suspend();
    }
    else {
      scout_swap_ipl(old_ipl);
    }
    
    xTrace0(sarp, TR_EVENTS, "sarRcv starting.");

    /*
     * dequeue messages from the rcv queue until the queue is empty
     */
    while ( (tail = sreg->rcvQTail) != sreg->rcvQHead ) {
      int  vci, len, bufIndex;
      char *virtAddr;
      u_int physAddr;
      Msg msg;

      xAssert((tail & ~(RCV_QUEUE_SIZE - 1)) == 0);

      /* get the message */
      bufIndex = sreg->rcvQueue[tail].bufIndex;
      len = sreg->rcvQueue[tail].count;
      vci = sreg->rcvQueue[tail].vci;
      xAssert( (bufIndex & ~(FREEB_QUEUE_SIZE - 1)) == 0);

      if (len == 0) {
	/*
	 * this means a packet was dropped
	 */
	xTrace0(sarp, TR_ERRORS, "sarRcv::dropping a packet\n");

	/* advance the queue tail */
	/*sreg->rcvQTail = (sreg->rcvQTail + 1) & (RCV_QUEUE_SIZE - 1); */
	dpmWriteInt(&(sreg->rcvQTail), (tail + 1) & (RCV_QUEUE_SIZE - 1)); 

	continue;
      }
      if (!(vci & 0x80000000)) {
	/*
	 * we don't support fragmented packets
	 */
	xTrace0(sarp, TR_ERRORS, "sarRcv::dropping a fragmented packet\n");

	/* advance the queue tail */
	/*sreg->rcvQTail = (sreg->rcvQTail + 1) & (RCV_QUEUE_SIZE - 1); */
	dpmWriteInt(&(sreg->rcvQTail), (tail + 1) & (RCV_QUEUE_SIZE - 1)); 

	continue;
      }

      virtAddr = soft->rcvFreeBuf[bufIndex].virtAddr;    

#ifdef IPHDR_CHECK
      if (ip_hdr_check &&
	  (*virtAddr != 0x46 ||
	   *(virtAddr+1) != 0x00 ||
	   *((short*)(virtAddr+2)) != 0x5a5a)) {
	printf("sarRcv: first word of header (pre-flush): ver/hlen=%x, type=%x, ident=%x\n",
	       *virtAddr, *(virtAddr+1), *((short*)(virtAddr+2)));
	
      }
#endif

      msgConstructInplace(&msg, virtAddr, 
			  MACHINE_PGBYTES, sarFreeRcvBuf);
#ifndef SIM_RCV_HACK
      msgTruncate(&msg, len);
#endif

      /* advance the queue tail */
      /*sreg->rcvQTail = (sreg->rcvQTail + 1) & (RCV_QUEUE_SIZE - 1); */
      dpmWriteInt(&(sreg->rcvQTail), (tail + 1) & (RCV_QUEUE_SIZE - 1)); 
    
    
#ifdef IPHDR_CHECK
      if (ip_hdr_check &&
	  (*virtAddr != 0x46 ||
	   *(virtAddr+1) != 0x00 ||
	   *((short*)(virtAddr+2)) != 0x5a5a)) {
	printf("sarRcv: first word of header: ver/hlen=%x, type=%x, ident=%x\n",
	       *virtAddr, *(virtAddr+1), *((short*)(virtAddr+2)));
	
      }
#endif

      /*
       * XXX--if the msg goes to user-land, need to bzero any trailing
       * unused portion, for security
       */

      xTrace2(sarp, TR_FULL_TRACE, 
	      "sarRcv:dequeued a frag, len=%d, vci=%x", len, vci);

      /*  add a fake eth header  */
      msgSetAttr(&msg, 0, &(soft->ethHdr), sizeof(ETHhdr));

      /* call demux to pass the msg up */
      xDemux(self, &msg);

      /*msgDestroy(&msg); done in ethDemux. YUCK!! */
    }
  }
}


/*
 * free a receive buffer
 * re-queue it on the free buffer queue
 */
void
  sarFreeRcvBuf(void* virtAddr, int len)
{
  SarSoftc *soft;
  SarReg *sreg;
  u_int physAddr;
  struct freeb_queue *freeBuf;
  int head, newHead;
  int bufIndex;

  soft = &sar_softc;  
  sreg = soft->sc_regs;

  bufIndex = ((char*)virtAddr - soft->rcvFreeBuf[0].virtAddr) / 
	MACHINE_PGBYTES;
  xAssert( (bufIndex & ~(FREEB_QUEUE_SIZE - 1)) == 0);  

  /* XXX - debug only */
  /*bzero(virtAddr, SARMTU);*/
  /*for (j=0; j<SARMTU/4; j++) 
    ((int*)virtAddr)[j] = 0x55aa55aa;*/

  /* 
   * re-queue the buffer 
   */
  head = sreg->freeBQHead;
  xAssert((head & ~(FREEB_QUEUE_SIZE - 1)) == 0);
#ifdef SIM_RCV_HACK
  newHead = head;
#else
  newHead = (head + 1) & (FREEB_QUEUE_SIZE - 1);
#endif
  /* queue should not be full */
  xAssert(newHead != sreg->freeBQTail);

  physAddr = soft->rcvFreeBuf[bufIndex].physAddr;
  freeBuf = &(sreg->freeBQueue[head]);

  /*sreg->freeBQueue[sreg->freeBQHead].bufIndex = bufIndex;*/
  dpmWriteInt(&(freeBuf->bufIndex), bufIndex);
  /*sreg->freeBQueue[sreg->freeBQHead].physAddr = physAddr;*/
  dpmWritePtr(&(freeBuf->physAddr), physAddr);
  /*sreg->freeBQueue[sreg->freeBQHead].size = SARMTU;*/
  dpmWriteInt(&(freeBuf->size), SARMTU);

  /* advance the free buffer queue head */
  /*sreg->freeBQHead = newHead;*/
  dpmWriteInt(&(sreg->freeBQHead), newHead);  
}



/***************************************/
/* Bottom half interrupt routines      */
/***************************************/

/* 
 * sar_intr()
 *
 */

void
sar_intr (XObj self)
{
  SarSoftc *soft;
  SarReg *sreg;
  int xmitCSR, rcvCSR;

  xTrace0(sarp, TR_EVENTS, "Interrupt handler for SAR.");

  soft = (SarSoftc*) self->state;
  if (!soft) {
    xTrace0(sarp, TR_ERRORS, "sar_intr: soft pointer is NULL.");
    return;
  }
  sreg = soft->sc_regs ;

  if ( (rcvCSR = sreg->rcvCSR) & SAR_INTR ) {
    /* 
     * receive SAR caused interrupt
     */
    xAssert( (rcvCSR & ~SAR_CSR_MASK) == 0);
  
    if ( rcvCSR & SAR_ERROR ) {
      /* possibly take some action hnere */
      xTrace0(sarp, TR_ERRORS, "sar_intr: Receive error condition bit set.") ;
    }

#ifdef TEST_HDR_INTR
    if ( rcvCSR & SAR_TEST ) {
      int lastBuf, bufIndex;
      char * virtAddr;

      /* verify the IP header */
      xTrace0(sarp, TR_EVENTS, "sar_intr: Testing IP header.");

      lastBuf = (sreg->freeBQTail - 1) & (FREEB_QUEUE_SIZE - 1);
      bufIndex = sreg->freeBQueue[lastBuf].bufIndex;
      virtAddr = soft->rcvFreeBuf[bufIndex].virtAddr;

      if (*virtAddr != 0x46 ||
	  *(virtAddr+1) != 0x00 ||
	  *((short*)(virtAddr+2)) != 0x5a5a) {
	
	/* header verification error */
	*((int *)sreg->rcvDump) = 0x5a5aa5a5;
	xTrace2(sarp, TR_ERRORS, "sar_intr: IP header error, pa=0x%x,val=0x%x", 
		soft->rcvFreeBuf[bufIndex].physAddr, *(int*)virtAddr);
      }

      /* clear SAR_INTR bit */
      /*sreg->rcvCSR = rcvCSR & ~SAR_INTR & ~SAR_PRDY;*/
      dpmWriteInt(&(sreg->rcvCSR), rcvCSR & ~SAR_INTR & ~SAR_TEST);
      
      rcvCSR &= ~SAR_TEST;
    }
#endif TEST_HDR_INTR
  
    /* if there is data ready from read SAR*/
    if ( rcvCSR & SAR_PRDY ) {

      /* wake a thread to execute sarRcv */
      scout_thread_async_wakeup(soft->rcvThread);
      
      /* clear SAR_INTR bit */
      /*sreg->rcvCSR = rcvCSR & ~SAR_INTR & ~SAR_PRDY;*/
      dpmWriteInt(&(sreg->rcvCSR), rcvCSR & ~SAR_INTR & ~SAR_PRDY);
    }
  }
  else if ( (xmitCSR = sreg->xmitCSR) & SAR_INTR ) {
    /* 
     * xmit SAR caused interrupt 
     */
    xAssert( (xmitCSR & ~SAR_CSR_MASK) == 0);
    xTrace0(sarp, TR_ERRORS, "sar_intr: Xmit interrupt, shouldn't happen.") ;
    
    if ( xmitCSR & SAR_ERROR ) {
      /* possibly take some action here */
      xTrace0(sarp, TR_ERRORS, "sar_intr: Xmit error condition bit set.") ;
    }
  }
  
  xTrace0(sarp, TR_EVENTS, "sar_intr exiting.") ;
}


/***************************************/
/* Random routines                     */
/***************************************/

/* 
 * sarReset()
 * reset the xmit and rcv side of the SAR board.
 * re-initialize queues.
 */
void
sarReset (XObj self)
{
  SarReg *sreg ;
  SarSoftc *soft;
  int i, nTries;

  xTrace0(sarp, TR_EVENTS, "sarReset.") ;
  soft = (SarSoftc*) self->state;
  if (!soft) {
    xTrace0(sarp, TR_ERRORS, "sarReset: soft pointer is NULL.");
    return;
  }
  sreg = soft->sc_regs ;

  /* 
   * reset both sar processors
   */
  semaLock(sreg, &sreg->xmitSema);
  dpmWriteInt(&sreg->xmitCSR, SAR_RESET); /* tell write processor to reset */
  semaUnlock(sreg, &sreg->xmitSema);
  
  semaLock(sreg, &sreg->rcvSema);
  dpmWriteInt(&sreg->rcvCSR, SAR_RESET); /* tell read processor to reset */
  semaUnlock(sreg, &sreg->rcvSema);

  /* wait for SAR_RDY */
  nTries = 0 ;
  while ( !(sreg->xmitCSR & SAR_RDY) ) {
    /* check to see if we've exceeded max number of retries*/
    if (nTries++ >= MAX_RETRY) {
      printf("sarReset: SAR Xmitter not responding.\n");
      break;
    }
    scout_udelay(5);
  }

  nTries = 0 ;
  while ( !(sreg->rcvCSR & SAR_RDY) ) {
    /* check to see if we've exceeded max number of retries*/
    if (nTries++ >= MAX_RETRY) {
	printf("sarReset: SAR Receiver not responding.\n");
	break;
    }
    scout_udelay(5);
  }

  /* re-initalize the receiver's free buffer queue */
  initRcvBQueue(soft);

}


    
/* 
 * lock the DPM semaphore
 */
int
semaLock(sreg, sema) 
SarReg *sreg;
volatile int *sema;
{
  int j, nTries = 0;

  xTrace1(sarp, TR_FULL_TRACE, "In semaLock, address of semaphore is %lx.", 
	  (u_long)sema) ;

  /* loop waiting for semaphore allocation */
  *sema = 0 ; /* ask for control of memory */
  scout_flush_write_to_memory();

  while ( (*sema & 0x000000ff) ) {
    if (++nTries > MAX_RETRY) {
      xTrace0(sarp, TR_ERRORS, "sema_lock:timeout");
      /*sarReset(0);*/
      return 0;
    }
    scout_udelay(5);
  }

  xTrace0(sarp, TR_FULL_TRACE, "semaLock: Got dual port, exiting.") ;
  return 1;
}

/* 
 * unlock the DPM semaphore
 */
void
semaUnlock(sreg, sema)
SarReg *sreg;
volatile int *sema;
{
  xTrace1(sarp, TR_FULL_TRACE, "In semaUnlock, address of semaphore is %lx.",
	  (u_long)sema) ;
  *sema = 0x000000ff ;
}



/*
 * Xmit stuff
 */

static void
  sarXmitUnwire( chPtr, chLen, arg )
char 	*chPtr;
long	chLen;
VOID    *arg;
{
#ifdef WIRE_XMIT_FBUFS
  fbUnwire(chPtr, chLen);
#endif
}


/*
 * XMIT_COLLECT_MAX_FRAMES limits the number of physical page frames per
 * message. 
 */
#define XMIT_COLLECT_MAX_FRAMES 32

typedef struct xmit_collect_desc {
  int    numFrags;
  SarBuf sarBufs[XMIT_COLLECT_MAX_FRAMES];
} XmitCollectDesc;


/*
 * apply functions for msgForEach in sarCtlrXmit
 */
static bool
  sarXmitCollect( virtAddr, len, arg )
char 	*virtAddr;
long	len;
VOID    *arg;
{
  XmitCollectDesc *xmitCD = (XmitCollectDesc*)arg;
  u_int physAddr;
  unsigned physLen, offset;
  char dum;

  xTrace2(sarp, TR_DETAILED, "sarXmitCollect, chunk at %lx, len=%d", 
	  (u_long)virtAddr, len);

#ifdef WIRE_XMIT_FBUFS
  fbWire(virtAddr, len);
#endif

  /*
   * need to translate each page individually
   * since pages may be non-contiguous in the physical address space
   */

  offset = (u_int)virtAddr & (MACHINE_PGBYTES - 1);
  dum = *(volatile char*)virtAddr;
  physAddr = kvtophys(virtAddr);
  
  if (len <= MACHINE_PGBYTES - offset) {
    /*
     * only one physical buffer 
     */
    physLen = len;
  }
  else {
    /*
     * more than one physical buffer
     */
    physLen = MACHINE_PGBYTES - offset;
    len -= physLen;
    virtAddr += physLen;

    while (1) {
      u_int nextPhysAddr;

      dum = *(volatile char*)virtAddr;
      nextPhysAddr = kvtophys(virtAddr);
      
      if (nextPhysAddr != physAddr + physLen) {
	/* 
	 * not contiguous, need to queue the previous buffer
	 */
	xmitCD->sarBufs[xmitCD->numFrags].physAddr = physAddr;
	xmitCD->sarBufs[xmitCD->numFrags].count = physLen;

	if (++(xmitCD->numFrags) >= XMIT_COLLECT_MAX_FRAMES) {
	  /* XXX - need to handle this  case gracefully */
	  xTrace0(sarp, TR_ERRORS, "sarXmitCollect: too many msg page frames");
	}

	physAddr = nextPhysAddr;
	physLen = 0;
      }

      if (len > MACHINE_PGBYTES) {
	virtAddr += MACHINE_PGBYTES;
	physLen += MACHINE_PGBYTES;      
	len -= MACHINE_PGBYTES;
      }
      else {
	physLen += len;
	break;
      }
    }
  }

  xmitCD->sarBufs[xmitCD->numFrags].physAddr = physAddr;
  xmitCD->sarBufs[xmitCD->numFrags].count = physLen;

  if (++(xmitCD->numFrags) >= XMIT_COLLECT_MAX_FRAMES) {
    /* XXX - need to handle this  case gracefully */
    xTrace0(sarp, TR_ERRORS, "too many msg page frames in sarXmitCollect");
  }
  
  return TRUE;
}


/*
 * sarCltrXmit
 *
 * transmit an outgoing packet
 *
 */

#ifdef SIM_RCV_HACK
int simNextRcvBuf = 0;
#endif

int
  sarCtlrXmit(downMsg, vci, seqNum)
Msg *downMsg;
int vci;
int  *seqNum;
{
  XmitCollectDesc xmitCD;
  SarSoftc *soft ;
  SarReg *sreg ;
  int head, nextHead;
  int k, i, j, nTries;

  xTrace1(sarp, TR_EVENTS, "sarCtlrXmit: sending a packet using VCI %d", vci);

  /* Init soft pointer */
  soft = &sar_softc;
  if (!soft) {
    xTrace0(sarp, TR_ERRORS, "sarCtlrXmit: soft pointer is NULL.\n") ;
    return 0;
  }
  sreg = soft->sc_regs;

  /* if no data return */
  if (msgLen(downMsg) == 0) {
    return 1;
  }

  /*
   * collect the msg's physical buffers 
   */
  xmitCD.numFrags = 0;

  msgForEach(downMsg, sarXmitCollect, &xmitCD);

  xTrace2(sarp, TR_EVENTS, "sarCtlrXmit: %d chunks, total bytes is %d", 
	  xmitCD.numFrags, msgLen(downMsg));

  /*
   * queue the buffers for transmission
   */
  for (i = 0; i < xmitCD.numFrags; i++) {
    struct xmit_queue *xmitDesc;

    /* queue next fragment */
    nTries = 0;

    /* xmitQHead won't change, so cache it in a local */
    head = sreg->xmitQHead;

    xAssert( (head & ~(XMIT_QUEUE_SIZE - 1)) == 0);
    nextHead = (head + 1) & (XMIT_QUEUE_SIZE - 1);

    while( nextHead == sreg->xmitQTail ) {
      /* 
       * xmitQueue is full, wait for SAR to free some space
       */
      if (++nTries > 50000) {
	/* timeout */
	xTrace2(sarp, TR_ERRORS, "sarCtlrXmit:xmitQueue timeout, frag %d of %d",
		i, xmitCD.numFrags);
	
	/* show fragments */
	for (j = 0; j < xmitCD.numFrags; j++) {
	  printf("frag %d at pa=%x, len %d\n", j, xmitCD.sarBufs[j].physAddr, 
		 xmitCD.sarBufs[j].count); 
	}

	return 0;
      }

      scout_udelay(100);
    }

    /* 
     * queue the fragment at xmitQueue[head] 
     */
    if (i == xmitCD.numFrags - 1) {
      /* 
       * last fragment of the packet
       * set the "last fragment" bit 
       */
      vci |= SAR_LAST_FRAG;

      /* 
       * save a copy of the msg, so it won't go away when users deallocate it. 
       * but first, un-wire the old xmit buffers associated with xmitMsg[head]
       */
#ifdef WIRE_XMIT_FBUFS
      if (msgLen(&(soft->xmitMsg[head])) != 0)
	msgForEach(&(soft->xmitMsg[head]), sarXmitUnwire, NULL);
#endif
      msgAssign(&(soft->xmitMsg[head]), downMsg);
    }

    xmitDesc = &(sreg->xmitQueue[head]);

#ifdef SEQNUM
    /*sreg->xmitQueue[head].seqNum = *seqNum;*/
    dpmWriteInt(&(xmitDesc->seqNum), *seqNum);
#endif

    /*xmitDesc->vci = vci;*/
    dpmWriteInt(&(xmitDesc->vci), vci);

    /*xmitDesc->count = xmitCD.sarBufs[i].count;*/
    dpmWriteInt(&(xmitDesc->count), xmitCD.sarBufs[i].count);
    
    /*xmitDesc->physAddr = xmitCD.sarBufs[i].physAddr;*/
    dpmWritePtr(&(xmitDesc->physAddr), xmitCD.sarBufs[i].physAddr);

    /* advance the xmitQHead */
    /*sreg->xmitQHead = nextHead;*/
    dpmWriteInt(&(sreg->xmitQHead), nextHead);

#ifdef SEQNUM
    /* update the cell seqNum */
    *seqNum = ( *seqNum + 
	        (xmitCD.sarBufs[i].count + (MAX_CELL_LEN - 1)) / MAX_CELL_LEN )
               & (CELL_SEQNUM_RANGE - 1); 
#endif

#ifdef SIM_RCV_HACK
    {
    static int hdrLen;

    if (i == 0) {
      /* copy the header into a receive buffer */
      printf("copying hdr with len=%d from addr=%p to addr=%p\n",
	     xmitCD.sarBufs[i].count,
	     (u_long)P_TO_K0SEG(xmitCD.sarBufs[i].physAddr),
	     (u_long)soft->rcvFreeBuf[simNextRcvBuf].virtAddr);

      bcopy(P_TO_K0SEG((char*)xmitCD.sarBufs[i].physAddr),
	    soft->rcvFreeBuf[simNextRcvBuf++].virtAddr,
	    xmitCD.sarBufs[i].count);

      hdrLen = xmitCD.sarBufs[i].count;
    }
    if (i == 1) {
      /* copy the msg body into a receive buffer */
      printf("copying body with len=%d from addr=%p to addr=%p\n",
	     xmitCD.sarBufs[i].count,
	     (u_long)P_TO_K0SEG(xmitCD.sarBufs[i].physAddr),
	     (u_long)soft->rcvFreeBuf[simNextRcvBuf - 1].virtAddr + hdrLen);

      bcopy(P_TO_K0SEG((char*)xmitCD.sarBufs[i].physAddr),
	    soft->rcvFreeBuf[simNextRcvBuf - 1].virtAddr + hdrLen,
	    xmitCD.sarBufs[i].count);
    }
    }
#endif

    xTrace2(sarp, TR_DETAILED, 
	    "sarCtlrXmit, queued a fragment at physAddr=%x, len=%d", 
	    xmitCD.sarBufs[i].physAddr, xmitCD.sarBufs[i].count);
  }

  xTrace0(sarp, TR_EVENTS, "sarCtlrXmit exiting.");
  return 1;
}


/*
 * read statistic from the DPM
 */
int
  sarGetStatistics(int *xmitStats, int* rcvStats)
{
  SarSoftc *soft ;
  SarReg *sreg ;
  int i;
 
  xTrace0(sarp, TR_EVENTS, "sarGetStatistics");

  /* Init soft pointer */
  soft = &sar_softc;
  if (soft == NULL) {
    xTrace0(sarp, TR_ERRORS, "sarGetStatistics: soft pointer is NULL.\n") ;
    return 0;
  }

  sreg = soft->sc_regs;

  xmitStats[0] = sreg->xmitIdles;
  xmitStats[1] = sreg->xmitFull;
  xmitStats[2] = sreg->xmitEmpty;
  
  rcvStats[0] = sreg->rcvCRCErrors;
  rcvStats[1] = sreg->rcvBQueueUnderflows; 
  rcvStats[2] = sreg->rcvQueueOverflows;
  rcvStats[3] = sreg->rcvOutOfOrderCells;
  rcvStats[4] = sreg->rcvOutOfOrderFrags;
  rcvStats[5] = sreg->rcvIdles;
  rcvStats[6] = sreg->rcvHdrFIFOFull;
  rcvStats[7] = sreg->rcvDMAFIFOFull;
  
  return 1;
}
