/*
 * $RCSfile: usc_main.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.2 $
 * $Date: 1996/02/15 18:42:09 $
 */

#include "usc.h"
#include "symbol.h"
#include "buffer.h"
#include "usc_validate.h"
#include "preprocess.h"
#include "gen_utils.h"
#include "usc_version.h"

#ifndef OPTIMIZE
extern void print_type_table (int level);
extern void print_sym_table (int level);
extern void print_struct_table (int level);
extern void print_union_table (int level);

#endif

extern void check_config (void);
extern void print_config (int level);
extern void init_symbols (void);
extern void init_codegen (void);
extern void finish_all_syms (void);
extern int yydebug;
extern FILE *yyin;
extern FILE *CodeOutFile;

static char *HelpMsg[] = {
  "Command line options:\n",
  "-D debug_level      -- select amount of debug output",
  "-V                  -- print version",
  "-h                  -- print this message",
  "-c                  -- stick comments in the generated code",
  "-p                  -- preserve type names in output",
  "-i filename         -- generate header file for prototypes and macros",
  "-s                  -- generate code to benchmark stubs",
  "-v                  -- generate code to run validation",
#ifdef YYDEBUG
  "-t                  -- print out yacc trace output",
#endif
  "-o filename         -- select output file",
  "",
  "Default options are: -o out.c ",
  NULL
};

char *allocate_new_string (char *string)
{
  char *new;
  int temp;
  if (string != NULL) {
    temp = strmem (string);
    new = (char *)Malloc (temp * sizeof (char));
    dprintf (1000, "Entering allocate_new_string: %s", string);
    strcpy (new, string);
    dprintf (1000, "Exiting allocate_new_string: %s %s", new, string);
  }
  else {
    new = NULL;
  }
  return new;
}

/***************************************************************************/
/*                                                                         */
/*  Here's some stubs for functions not used in USIT but not by USC        */
/*                                                                         */
/***************************************************************************/


void usit_add_function (Buffer *source)
{
}

/***************************************************************************/

char *allocate_temp_string (char *string)
{
  char *new;
  if (string != NULL) {
    new = (char *)TempMalloc (strmem (string) * sizeof (char));
    dprintf (1000, "Entering allocate_temp_string: %s", string);
    strcpy (new, string);
    dprintf (1000, "Exiting allocate_temp_string: %s %s", new, string);
  }
  else {
    new = NULL;
  }
  return new;
}

int strmem (char *string)
{
  return strlen (string) + 1;
}

void init (void)
{
  init_mem();
  inityy();
  init_symbols();
  init_gen_utils(OutFile);
  init_codegen();
  InFileName = NULL;
  OutFile = NULL;
  HeaderFile = NULL;
  DefaultCodeType = 0;
  code_type = DefaultCodeType;
  print_comments = 0;
  make_validate = 0;
  speed_test = 0;
#ifdef USE_DMALLOC
  dmalloc_opt (DMALLOC_HEAPSTAT | DMALLOC_HEAPCHECK); 
#endif
}

void finishup(void)
{
  static int done = 0;

  if ( done )
    return;
  else
    done = 1;

  if (make_validate)
    close_validate();

  /*
  if (preserve_type_names)
    write_header_file(HeaderFile);
  */
}

void closeup (void)
{
  finishup();

  close_mem();
#ifdef USE_DMALLOC
  print_heap_stats();
#endif
  if (OutFile != stdout)
    fclose (OutFile);
  if (HeaderFile && HeaderFile != OutFile)
    fclose(HeaderFile);
}

int main(int argc, char **argv)
{
  extern char* optarg;
  extern int   optind;
  int          option;
  char         *options;
  char *progname = argv[0];
  char *out_file_name;

#ifdef YYDEBUG
  options = "D:Vhco:pi:stv";
#else
  options = "D:Vhco:pi:sv";
#endif

  init();

  InFileName = NULL;
				/* The value of optind is undefined
                                   before getopt is called.  Here we
                                   force it to one so we can skip
                                   argv[0].  This was needed for
                                   Linux. */
  optind = 1;

  while (optind < argc) {
    if (argv[optind][0] == '-') {
      option = getopt(argc, argv, options);
      if (option == EOF) {
	break;
      }
      
      switch (option) {
      case 'D':
	Debug_Level = atoi (optarg);
	break;
      
      case 'V':
	{
	  char** vers;
	  for ( vers=Version; *vers; vers++ ) {
	    fprintf(stderr,"%s\n",*vers);
	  }
	  exit(0);
	}
	break;

      case 'h':
	{
	  char **help;
	  for (help = HelpMsg; *help; help++) {
	    fprintf (stderr, "%s\n", *help);
	  }
	  exit (0);
	}
	break;

      case 'c':
	print_comments = 1;
	break;

      case 'p':
	preserve_type_names = 1;
	break;
      
      case 'i':
	if ( HeaderFile ) {
	  fprintf(stderr, "%s: Two definitions for the header file\n",
		  progname);
	  exit(1);
	}
	HeaderFileName = allocate_new_string(optarg);
	HeaderFile = fopen (HeaderFileName, "w");
	if (!HeaderFile) {
	  fprintf (stderr, "%s: Can't open header file: %s\n", progname,
		   HeaderFileName);
	  exit (1);
	}
	break;

      case 's':
	make_validate = 1;
	speed_test = 1;
	print_comments = 0;
	break;

#ifdef YYDEBUG
      case 't':
	yydebug = 1;
	break;
#endif

      case 'v':
	make_validate = 1;
	print_comments = 0;
	break;

      case 'o':
	if ( OutFile ) {
	  fprintf(stderr, "%s: Two definitions for the output file\n",
		  progname);
	  exit(1);
	}
	out_file_name = allocate_new_string ( optarg );
	if (!strcmp (out_file_name, "-"))
	  OutFile = stdout;
	else
	  OutFile = fopen (out_file_name, "w");
	if (!OutFile) {
	  fprintf (stderr, "%s: Can't open output file: %s\n", progname,
		   out_file_name);
	  exit (1);
	}
	break;
      }
    }
    else {
      if (InFileName) {
	fprintf (stderr, "%s: Can't have two input files\n", progname);
	exit (1);
      }
      else {
	InFileName = allocate_new_string (argv[optind]);
	if (strcmp (InFileName, "-")) {
	  yyin = fopen (InFileName, "r");
	  if (!yyin) {
	    fprintf (stderr, "%s: Can't open input file: %s\n", progname,
		     argv[optind]);
	    exit (1);
	  }
	}
      }
      optind++;
    }
  }
  argv += optind;
  argc -= optind;


  if ( argc > 1 ) {
    fprintf( stderr, "%s: Unknown option\n", progname);
    exit (1);
  }

  if ( !InFileName ) {
    InFileName = "-";
  }

  if ( !OutFile ) {
    if (make_validate)
      out_file_name = allocate_new_string( VALIDATE_FILE );
    else
      out_file_name = allocate_new_string( OUTPUT_FILE );
    OutFile = fopen (out_file_name, "w");
    if (!OutFile) {
      fprintf (stderr, "%s: Can't open output file: %s\n", 
	       progname, out_file_name);
      exit (1);
    }
  }

  if ( !HeaderFile )
    HeaderFile = OutFile;

#if 0
  if ( preserve_type_names && !HeaderFile ) {
    char header_file_name[128];
    char *suffix;

    strcpy(header_file_name,out_file_name);
    suffix = strstr(header_file_name,".c");
    if ( !suffix || suffix != header_file_name+(strlen(header_file_name)-2) ) {
      fprintf (stderr, "Type descriptions in file %s.h\n",header_file_name);
    }
    else {
      *suffix = '\0';
    }
    strcat(header_file_name,".h");
    HeaderFileName = allocate_new_string(header_file_name);
    HeaderFile = fopen (HeaderFileName, "w");
    if (!HeaderFile) {
      fprintf (stderr, "%s: Can't open type description file: %s\n", 
	       progname, header_file_name);
      exit (1);
    }
  }
#endif

  init_outfile( USC_OUTPUT );
  if ( HeaderFile && HeaderFile != OutFile ) {
    init_outfile( USC_TYPES );
  }

  yyparse();

  print_sym_table (200);
  print_type_table (200);
  
  print_struct_table (250);
  print_union_table (250);

  /*
  if (make_validate)
    close_validate();

  if (preserve_type_names)
    write_header_file(HeaderFile);
  */
  
  closeup();
  return 0;
}
