/*
 * $RCSfile: usit_gen.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/13 17:43:30 $
 */

#include "usc.h"
#include "usit.h"
#include "gen_utils.h"
#include "symbol.h"
#include "bit_twiddle.h"

typedef enum { INPUT = 1, OUTPUT = 0 } PreFlag;

typedef struct {
  int byte_offset;
  int bit_offset;
} BitfieldMap;

static FILE *pre_pipe;
static int which_pass;
static Buffer *UsitInferMain;

#ifndef MAX_ALIGN
#define MAX_ALIGN 8
#endif

extern void declare_sym (int num_tabs, Symbol *sym, FILE *out);

char *pragma_local_name (BucketInfo *info)
{
  static char name[1024];

  sprintf (name, "usit_pragma_%s", info->name);

  return name;
}

char *align_struct_name (char *symname, int padding)
{
  char *prefix = "usit_align";
  char *name;
  char num[4];

  sprintf (num, "%d", padding);
  
  name = (char *)Malloc ((strlen (prefix) + strlen (symname) + 
			  strlen (num) + 2) *
			 sizeof (char));
  sprintf (name, "%s%s_%s", prefix, num, symname);
  
  return name;
}

void send_flags_to_brothers (Type *type, PreFlag flag)
{
  Type *p = type;

  do {
    TYPE_USIT_SEEN (p, flag) = TYPE_USIT_SEEN (type, flag);
    p = TYPE_NEXTBRO (p);
  } while (p != type);
}

void update_all_brothers (TypeData *data, Type *type)
{
  Type *p = type;
  int offset;

  do {
    if (TYPE_MODIFIERS(p)) {
      offset = TYPEDATA_OFFSET (TYPE_MODIFIERS (p));
      *TYPE_MODIFIERS(p) = *data;
      if (p != type)
	TYPEDATA_OFFSET (TYPE_MODIFIERS (p)) = offset; 
    }
    else {
      TYPE_MODIFIERS(p) = (TypeData *)Malloc (sizeof (TypeData));
      *TYPE_MODIFIERS(p) = *data;
      if (p != type)
	TYPEDATA_OFFSET (TYPE_MODIFIERS (p)) = DEFAULT_OFFSET;
    }
    p = TYPE_NEXTBRO (p);
  } while (p != type);
}
 
char *make_temp_file (void)
{
#ifdef NO_TEMPNAM
  char *name = Malloc(16);
  sprintf(name,"./usc%06d",getpid());
#else
  char *name = tempnam (".", TEMP_PREFIX);
#endif
  
  dprintf (10, "temp file: %s", name);

  return name;
}

int get_next_pipe_int (int *value)
{
  int flag;
  flag = fscanf (pre_pipe, "%d", value);
  if (flag == EOF)
    crash ("internal error in get_next_pipe_int");
  return flag;
}

int get_next_pipe_hexint ( unsigned long * value)
{
  int flag;
  flag = fscanf (pre_pipe, "%lx", value);
  if (flag == EOF)
    crash ("internal error in get_next_pipe_hexint");
  return flag;
}

int get_next_pipe_border (char *string)
{
  int val;
  int flag;
  char* border = string;

  while ( (flag = get_next_pipe_int(&val)) && val )
    if ( val > MAX_BORDER_SIZE )
      crash("internal error in get_next_pipe_border");
    else
      *(border++) = val;
  dprintf(50, "byte order: %s", byteorder_to_string(string) );
  return flag;
}

int get_next_pipe_string (char *string)
{
  int flag;
  flag = fscanf (pre_pipe, "%s", string);
  if (flag == EOF)
    crash ("internal error in get_next_pipe_string");
  dprintf (50, "string: %s", string);
  return flag;
}

Type *make_native (Type *thetype)
{
  if (TYPE_ADDNATIVE (thetype))
    crash ("Already declared native");
  if (!TYPE_NATIVE(thetype))
    crash ("Can't declare this as native");
  TYPE_ADDNATIVE (thetype) = 1;
  return thetype;
}

void init_preout_file (FILE *outfile)
{
  fprintf (outfile, "#include <stdio.h>\n\n");
}

UsitDecl *new_decl (UsitClass class)
{
  UsitDecl *new;

  new = (UsitDecl *)Malloc (sizeof (UsitDecl));
  new->the_class = class;
  new->the_sym = NULL;
  new->next = NULL;

  return new;
}

void init_dec_list (UsitDecList *list)
{
  list->head = NULL;
  list->tail = NULL;
}

void append_decl (UsitDecList *list, UsitDecl *decl)
{
  decl->next = NULL;

  if (list->head) {
    list->tail->next = decl;
    list->tail = decl;
  }
  else {
    list->head = decl;
    list->tail = decl;
  }
}

void print_dec_list (int level, UsitDecList *list)
{
  UsitDecl *p;

  if (level >= Debug_Level) {
    dprintf (level, "The Decl List:\n");
    for (p = list->head; p; p = p->next) {
      print_sym (level, 0, "", p->the_sym);
    }
  }
}

Symbol *struct_to_sym (char *prefix, Type *the_struct)
{
  Symbol *new;

  char *name = TYPE_STRUCTDEF_NAME (the_struct);
  char *new_name = (char *)Malloc (strlen (prefix) + strlen (name) + 1);
  sprintf (new_name, "%s%s", prefix, name);
  
  new = (Symbol *)Malloc (sizeof (Symbol));
  SYM_NAME(new) = allocate_new_string (name);
  SYM_LOCAL_NAME(new) = new_name;
  SYM_NEXT(new) = NULL;
  SYM_END(new) = new;
  SYM_TYPEDATA(new) = NULL;
  SYM_LEVEL(new) = 0;

  SYM_TYPE (new) = the_struct;
  SYM_ETYPE (new) = TYPE_ETYPE (the_struct);

  return new;
}

void declare_typedef (Symbol **sym, UsitDecList *the_list)
{
  UsitDecl *new;

  if (SYM_NATIVE (*sym)) {
    dprintf (20, "declaring typedef %s", SYM_NAME (*sym));
    new = new_decl (TYPEDEF_CLASS);
    new->the_sym = *sym;
    append_decl (the_list, new);
  }
}

void declare_struct (Type **type, UsitDecList *the_list)
{
  UsitDecl *new;

  if (TYPE_NATIVE (*type) && TYPE_STRUCTDEF_NAME (*type)) {
    dprintf (20, "declaring struct %s", TYPE_STRUCTDEF_NAME (*type));
    new = new_decl (STRUCT_CLASS);
    new->the_sym = struct_to_sym ("usit_struct_", *type);
    append_decl (the_list, new);
  }
}

void declare_union (Type **type, UsitDecList *the_list)
{
  UsitDecl *new;
  char *shit;

  if (TYPE_NATIVE (*type) && TYPE_STRUCTDEF_NAME (*type)) {
    shit = TYPE_STRUCTDEF_NAME (*type);
    dprintf (20, "declaring union %s", shit);
    new = new_decl (UNION_CLASS);
    new->the_sym = struct_to_sym ("usit_union_", *type);
    append_decl (the_list, new);
  }
}

void build_dec_list (UsitDecList *the_list)
{
  ptab (Type_Tab, declare_typedef, the_list, 0);
/*  ptab (Struct_Defs, declare_struct, the_list, 0);
  ptab (Union_Defs, declare_union, the_list, 0);
*/
}

void print_struct (int num_tabs, Type *spec, FILE *out)
{
  Symbol *sym;
  for (sym = TYPE_STRUCTDEF_FIELDS(spec); sym; sym = SYM_NEXT(sym)) {
    declare_sym (num_tabs + 1, sym, out);
  }
}

void print_specifiers (int num_tabs, Symbol *sym, FILE *out)
{
  extern char *spec_type_to_string (Type* type);
  char *indent = make_indent (num_tabs);
  SpecifierTypeField specfield;

  if (!TYPE_IS_SPECIFIER (SYM_ETYPE(sym)))
    crash ("internal error in preout_specifiers");
  specfield = TYPE_SPEC_TYPE (SYM_ETYPE (sym));
  fprintf (out, "%s%s", indent, spec_type_to_string (SYM_ETYPE (sym)));
  if (specfield & (UNION_TYPE | STRUCT_TYPE)) {
    fprintf (out, "{\n");
    print_struct (num_tabs, SYM_ETYPE(sym), out);
    fprintf (out, "%s} ", indent);
  }
  else if (specfield & ENUM_TYPE) {
    crash ("no stinking enums wanted here");
  }
  if (indent[0])
    Free (indent, 0);
}

void print_name (int num_tabs, Symbol *sym, FILE *out)
{
  if (which_pass == 0) {
    fprintf (out, "%s", SYM_LOCAL_NAME (sym));
  }
  else {
    fprintf (out, "%s", SYM_NAME (sym));
  }
}

int get_num_elements (Type *thetype)
{
  int ans = 1;

  if (TYPE_DECL_NUMEL (thetype) >= 0) {
    while (TYPE_IS_DECLARATOR (thetype)) {
      ans *= TYPE_DECL_NUMEL (thetype);
      thetype = TYPE_NEXT (thetype);
    }
  }

  return ans;
}

void print_modifiers (int num_tabs, Symbol *sym, FILE *out)
{
  char *border = "0";
  int  tsize = 0, memsize, offset, alignment;

  memsize = SYM_MEMSIZE(sym);
  if (SYM_IS_ARRAY(sym)) {
    memsize /= get_num_elements (SYM_TYPE (sym));
  }
  offset  = SYM_OFFSET (sym);
  alignment = SYM_ALIGN(sym);
  if ( !SYM_IS_STRUCT_OR_UNION(sym) ) {
    tsize = SYM_SIZE(sym);
    if ( SYM_BORDER (sym) )
      border = byteorder_to_string( SYM_BORDER (sym) );
  }

  if (num_tabs)
    fprintf (out, "(%d, %d, %d, %s)", tsize, memsize, offset, border);
  else
    fprintf (out, "(%d, %d, %d, %s)", tsize, memsize, alignment, border);
}

void print_declarators (Symbol *sym, FILE *out)
{
  Type *type = SYM_TYPE (sym);
  while (type && !TYPE_IS_SPECIFIER(type)) {
    if (TYPE_IS_ARRAY (type)) {
      if (TYPE_DECL_NUMEL (type) >= 0) {
	fprintf (out, "[%d]", TYPE_DECL_NUMEL (type));
      }
      else {
	if (which_pass == 0)
	  fprintf (out, "[1]");
	else
	  fprintf (out, "[%s]", EXPR_IDENT (TYPE_DECL_VAR_ARRAY (type)));
      }
    }
    else if (TYPE_IS_BITFIELD (type)) {
      if (which_pass == 0) {
	fprintf (out, " : %d", TYPE_DECL_NUMEL (type));
      }
      else {
	fprintf (out, " : %d(%d,%d,%d,%s)", TYPE_DECL_NUMEL (type),
		 TYPE_DECL_SIZE (type), TYPE_DECL_MEMSIZE (type), 0,
		 TYPE_DECL_BITINFO (type) ? 
		 bitorder_to_string (TYPE_DECL_BITINFO (type)) : "<zippo>");
      }
    }
    else {
      crash ("internal error in preout_declarators");
    }
    type = TYPE_NEXT(type);
  }
}

void declare_sym (int num_tabs, Symbol *sym, FILE *out)
{
  print_specifiers (num_tabs, sym, out);
  print_name (num_tabs, sym, out);
  if (which_pass == 1)
    print_modifiers (num_tabs, sym, out);
  print_declarators (sym, out);
  fprintf (out, ";\n");
}  

void output_align_tester (Symbol *sym, int padding, FILE *out)
{
  char *struct_name;
  char *indent = make_indent (1);
  
  struct_name = align_struct_name (SYM_LOCAL_NAME (sym), padding);
  
  fprintf (out, "struct {\n");
  if (padding)
    fprintf (out, "%schar usit_padchar[%d];\n", indent, padding);
  declare_sym (1, sym, out);
  fprintf (out, "} %s;\n\n", struct_name);

  if (indent[0])
    Free (indent, 0);
  Free (struct_name, strmem (struct_name));
}

void output_pragma_align_tester (BucketInfo *info, int padding, FILE *out)
{
  char *struct_name;
  char *indent = make_indent (1);
  
  struct_name = align_struct_name (pragma_local_name (info), padding);

  fprintf (out, "struct {\n");
  if (padding)
    fprintf (out, "%schar usit_padchar[%d];\n", indent, padding);
  fprintf (out, "%s%s %s;\n", indent, info->name, pragma_local_name (info));
  fprintf (out, "} %s;\n\n", struct_name);

  Free (indent, 0);
  Free (struct_name, 0);
}

void output_sym_declarations (Symbol *sym, FILE *out)
{
  int i;
  for (i = 1; i <= MAX_ALIGN; i++) {
    output_align_tester (sym, i, out);
  }
  declare_sym (0, sym, out);
}

void declare_pragma (BucketInfo *info, FILE *out)
{
  fprintf (out, "%s %s;\n\n", info->name, pragma_local_name (info));
}

void declare_pragma_align_testers (BucketInfo *info, FILE *out)
{
  int i;

  for (i = 1; i <= MAX_ALIGN; i++) 
    output_pragma_align_tester (info, i, out);
}

void declare_pragmas (FILE *out)
{
  if (machine_info.char_info.size == DEFAULT_SIZE) {
    declare_pragma (&machine_info.char_info, out);
    declare_pragma_align_testers (&machine_info.char_info, out);
  }
  if (machine_info.short_info.size == DEFAULT_SIZE) {
    declare_pragma (&machine_info.short_info, out);
    declare_pragma_align_testers (&machine_info.short_info, out);
  }
  if (machine_info.int_info.size == DEFAULT_SIZE) {
    declare_pragma (&machine_info.int_info, out);
    declare_pragma_align_testers (&machine_info.int_info, out);
  }
  if (machine_info.long_info.size == DEFAULT_SIZE) {
    declare_pragma (&machine_info.long_info, out);
    declare_pragma_align_testers (&machine_info.long_info, out);
  }
}

void declare_everything (UsitDecList *the_list, FILE *out)
{
  UsitDecl *p;

  declare_pragmas (out);

  for (p = the_list->head; p; p = p->next) {
    output_sym_declarations (p->the_sym, out);
    fprintf (out, "\n");
  }
}

int pipe_alignment (char *symname, PreFlag flag)
{
  int alignment = 0, i, temp;
  char *struct_name;

  switch (flag) {
  case INPUT:
    for (i = 1; i <= MAX_ALIGN; i++) {
      get_next_pipe_int (&temp);
      alignment = gcd (temp, alignment);
    }
    dprintf (20, "%s alignment: %d", symname, alignment);
    break;

  case OUTPUT:
    for (i = 1; i <= MAX_ALIGN; i++) {
      struct_name = align_struct_name (symname, i);
      
      writebuf (func_body, "printf (\"%%d \", (int)&(%s.%s) - (int)&%s);\n",
		struct_name, symname, struct_name);

      Free (struct_name, strmem (struct_name));
    }
    writebuf (func_body, "printf (\"\\n\");\n");
    break;

  default:
    crash ("internal error in figure_alignment");
  }

  return alignment;
}

void pipe_size_offset_code (char *basename, char *name,
			    int last_offset, TypeData *data, PreFlag flag)
{
  int size, offset;

  switch (flag) {
  case INPUT:
    if (data == NULL) {
      crash ("internal error in pipe_size_offset_code");
    }
    get_next_pipe_int (&size);
    get_next_pipe_int (&offset);
    TYPEDATA_SIZE (data) = size;
    TYPEDATA_MEMSIZE (data) = size;
    TYPEDATA_OFFSET (data) = offset - last_offset;
    dprintf (50, "%s size: %d offset: %d - %d", name, size, offset, 
	     last_offset);
    break;

  case OUTPUT:
    writebuf (func_body, "printf (\"%%d %%d\\n\", sizeof (%s), "
	      "(int)(&%s) - (int)(&%s));\n", name, name, 
	      basename);
    break;
  }
}  

BucketInfo *guess_bitfield_basetype (int min, int max)
{
  int size = max - min + 1;
  BucketInfo *cur_info, *best_info = NULL;
  
  /* find smallest enclosing type. As long as the size is cool, don't 
     worry about the alignment, cause usc doesn't care */

  cur_info = &(machine_info.long_info);
  if (cur_info->size >= size) {
    if (best_info == NULL || best_info->size > cur_info->size)
      best_info = cur_info;
  }
  cur_info = &(machine_info.int_info);
  if (cur_info->size >= size) {
    if (best_info == NULL || best_info->size > cur_info->size)
      best_info = cur_info;
  }
  cur_info = &(machine_info.short_info);
  if (cur_info->size >= size) {
    if (best_info == NULL || best_info->size > cur_info->size)
      best_info = cur_info;
  }
  cur_info = &(machine_info.char_info);
  if (cur_info->size >= size) {
    if (best_info == NULL || best_info->size > cur_info->size)
      best_info = cur_info;
  }

  return best_info;
}

BitInfo *figure_bit_info (BucketInfo *bucket, int min_offset, int max_offset, 
			  BitfieldMap *bit_map, int num_bits)
{
  BitInfo *info = (BitInfo *)Malloc (MAX_BITFIELD_SIZE * sizeof (BitInfo));
  int i;

  for (i = 0; i < MAX_BITFIELD_SIZE; i++)
    info[i] = -1;

  for (i = 0; i < num_bits; i++) {
    info[i] = pos_to_bit (bucket->byteorder, 
			  (bit_map[i].byte_offset - min_offset) * 8 + 
			  (7 - bit_map[i].bit_offset));
  }

  return info;
}

void pipe_figuring_code (Symbol *basesym, Type *type, char *name,
			 int last_offset, TypeData *data, PreFlag flag)
{
  char border[128] = { '\0' };
  int i, j, offset, num_bits, min_offset, max_offset;
  unsigned long bit;
  char c;
  BitfieldMap *bit_map;
  BucketInfo *bucket;
  BitInfo *bit_info;

  switch (flag) {
  case INPUT:
    if (data == NULL) {
      crash ("internal error in pipe_figuring_code");
    }
    if (TYPE_IS_BITFIELD (type)) {
      num_bits = TYPE_DECL_NUMEL (type);
      bit_map = (BitfieldMap *)Malloc (sizeof (BitfieldMap) * num_bits);
      for (i = 0; i < num_bits; i++) {
	get_next_pipe_int (&offset);
	get_next_pipe_hexint (&bit);
	if (num_set_bits (bit) != 1)
	  crash ("Big problem in pipe_figuring_code!");
	bit = which_bit_set (bit);
	bit_map[i].byte_offset = offset;
	bit_map[i].bit_offset = bit;
      }
      min_offset = INT_MAX;
      max_offset = INT_MIN;
      for (i = 0; i < num_bits; i++) {
	dprintf (20, "%s bit %d is at (%d, %d)", name, i, 
		 bit_map[i].byte_offset, bit_map[i].bit_offset);
	min_offset = MIN (min_offset, bit_map[i].byte_offset);
	max_offset = MAX (max_offset, bit_map[i].byte_offset);
      }
      dprintf (20, "%s min: %d max: %d", name, min_offset, max_offset);
      bucket = guess_bitfield_basetype (min_offset, max_offset);
      if (bucket == NULL) {
	error ("Severe badness. USIT failed to figure out bitfield");
	return;
      }
      dprintf (20, "%s should use %s", name, bucket->name);
      TYPEDATA_SIZE (data) = bucket->size;
      TYPEDATA_MEMSIZE (data) = bucket->memsize;
      TYPEDATA_OFFSET (data) = min_offset - last_offset;
      TYPEDATA_BORDER (data) = bucket->byteorder;

      bit_info = figure_bit_info (bucket, min_offset, max_offset, 
				  bit_map, num_bits);
      
      TYPE_DECL_BITINFO (type) = bit_info;
      TYPE_DECL_SIZE (type) = num_bits;
      TYPE_DECL_MEMSIZE (type) = 8 * bucket->memsize;

      free (bit_map);
    }
    else if (TYPE_IS_SIMPLE (type)) {
      pipe_size_offset_code (SYM_LOCAL_NAME (basesym), name, 
			     last_offset, data, flag);
      get_next_pipe_border(border);
      for (i = 0, j = strlen(border) - 1; i < j; i++, j--) {
	c = border[i];
	border[i] = border[j];
	border[j] = c;
      }
      TYPEDATA_BORDER (data) = allocate_new_string (border);
      dprintf (50, "%s border: %s", name, byteorder_to_string (border));
    }
    else {
      crash ("internal error in pipe_figuring_code");
    }
    break;

  case OUTPUT:
    if (TYPE_IS_BITFIELD (type)) {
      char *basename = SYM_LOCAL_NAME (basesym);

      writebuf (func_body, 
		"for (j = 0; j < %d; j++) ", TYPE_DECL_NUMEL (type));
      push_code_level();
      writebuf (func_body, 
		"for (i = 0; i < sizeof (%s); i++) "
		"*((unsigned char *)(&%s)+i) = \'\\0\';\n", 
		basename, basename);
      writebuf (func_body, "%s = 1L << j;\n", name);
      writebuf (func_body, "for (i = 0; i < sizeof (%s); i++) ",basename);
      push_code_level();
      writebuf (func_body, "if (*((unsigned char *)(&%s)+i)) ", basename);
      push_code_level();
      writebuf (func_body, 
		"printf (\"%%d %%x\\n\", i, *((unsigned char *)(&%s)+i));\n",
		basename);
      pop_code_level();
      pop_code_level();
      pop_code_level();
    }
    else if (TYPE_IS_SIMPLE (type)) {
      pipe_size_offset_code (SYM_LOCAL_NAME (basesym), name, 
			     last_offset, data, flag);
      writebuf (func_body, "for (i = 0; i < sizeof (%s); i++) ", name);
      push_code_level();
      writebuf (func_body, "%s <<= 8;\n", name);
      writebuf (func_body, "%s |= i + 1;\n", name);
      pop_code_level();
      writebuf (func_body, "cptr = (char *)&(%s);\n", name);
      writebuf (func_body, "for (i = 0; i < sizeof (%s); i++) ", name);
      push_code_level();
      writebuf (func_body, "printf (\"%%d \", *(cptr+i));\n");
      pop_code_level();
      writebuf (func_body, "printf (\"0\\n\");\n");
    }
    else {
      crash ("internal error in pipe_figuring_code");
    }
    break;
  }
}

void precode_sym (Symbol *basesym, Symbol *sym, char *prefix, int last_offset,
                  int alignment, PreFlag flag) 
{
  Symbol *field;
  int i, offset;
  Type *type;
  TypeData *data;
  char *name = NULL, *string = NULL;
  int name_len, string_len;
  UsitTypeSeen *usit_seen;

  if (flag == INPUT) {
    data = make_default_typedata();
    if (TYPE_MODIFIERS (SYM_ETYPE (sym)) == NULL) 
      TYPE_MODIFIERS (SYM_ETYPE (sym)) = data; 
  }
  else {
    data = NULL;
  }

  name_len = strlen (prefix) + strlen (SYM_LOCAL_NAME (sym));
  name = (char *)Malloc ((name_len + 1) * sizeof (char));
  sprintf (name, "%s%s", prefix, SYM_LOCAL_NAME (sym));
  
  usit_seen = &TYPE_USIT_SEEN (SYM_ETYPE (sym), flag);

  dprintf (50, "precode_sym: %s", name);

  if (TYPE_IS_SPECIFIER (SYM_TYPE (sym))) {
    if (SYM_IS_STRUCT_OR_UNION (sym)) {
      string_len = name_len + 1;

      string = (char *)Malloc ((string_len + 1) * sizeof (char));
      sprintf (string, "%s.", name);

      pipe_size_offset_code (SYM_LOCAL_NAME (basesym), name, last_offset, 
			     data, flag);
      if (data)
	offset = TYPEDATA_OFFSET (data) + last_offset;
      else
	offset = 0;

      if (*usit_seen == USIT_NOT_SEEN) {
	dprintf (10, "%s not seen (pass %d)", name, flag);
	for (field = TYPE_STRUCTDEF_FIELDS (SYM_TYPE (sym)); field; 
	     field = SYM_NEXT (field)) {
	  precode_sym (basesym, field, string, offset, -1, flag);
	}
      }
      else {
	dprintf (10, "%s already seen! (pass %d)", name, flag);
      }

      Free (string, string_len + 1);
    }
    else {
      pipe_figuring_code (basesym, SYM_TYPE (sym), name, last_offset, 
			  data, flag);
    }
  }
  else if (SYM_IS_ARRAY (sym)) {
    type = SYM_TYPE (sym);
    i = 0;
    while (TYPE_IS_ARRAY (type)) {
      i++;
      type = TYPE_NEXT (type);
    }
      
    string_len = name_len + 3 + i * 3;
    string = (char *)Malloc ((string_len + 1) * sizeof (char));
    sprintf (string, "%s", name);
    while (i) {
      strcat (string, "[0]");
      i--;
    }
    if (TYPE_IS_STRUCT_OR_UNION (type)) {
      pipe_size_offset_code (SYM_LOCAL_NAME (basesym), name, last_offset, 
			     data, flag);
      if (data)
	offset = TYPEDATA_OFFSET (data) + last_offset;
      else
	offset = 0;
      strcat (string, ".");
      if (*usit_seen == USIT_NOT_SEEN) {
	for (field = TYPE_STRUCTDEF_FIELDS (type); field; 
	     field = SYM_NEXT (field)) {
	  precode_sym (basesym, field, string, offset, -1, flag);
	}
      }
      if (data) {
	TYPEDATA_SIZE (data) /= get_num_elements (SYM_TYPE (sym));
	TYPEDATA_MEMSIZE (data) = TYPEDATA_SIZE (data);
      }
    }
    else {
      pipe_figuring_code (basesym, SYM_ETYPE (sym), string, last_offset, 
			  data, flag);
    }
    Free (string, string_len + 1);
  }
  else if (SYM_IS_BITFIELD (sym)) {
    pipe_figuring_code (basesym, SYM_TYPE (sym), name, last_offset, 
			data, flag);
  }
  else {
    crash ("internal error in precode_sym");
  }
      
  *usit_seen = USIT_SEEN;

  if (data) {
    if (alignment > 0)
      TYPEDATA_ALIGN (data) = alignment;
    /* update_all_brothers (data, SYM_ETYPE (sym));   */
  }

  send_flags_to_brothers (SYM_ETYPE (sym), flag);

  Free (name, name_len + 1);
}

void pipe_pragma (BucketInfo *info, PreFlag flag)
{
  int alignment, i, j;
  char name[1024], border[128] = { '\0' }, c;
  TypeData the_data;

  strcpy (name, pragma_local_name (info));

  switch (flag) {
  case INPUT:
    alignment = pipe_alignment (name, flag);
    pipe_size_offset_code (name, name, 0, &the_data, flag);
    get_next_pipe_border(border);
    for (i = 0, j = strlen(border) - 1; i < j; i++, j--) {
      c = border[i];
      border[i] = border[j];
      border[j] = c;
    }
    TYPEDATA_BORDER (&the_data) = allocate_new_string (border);

    info->size = TYPEDATA_SIZE (&the_data);
    info->memsize = TYPEDATA_MEMSIZE (&the_data);
    info->alignment = alignment;
    info->byteorder = TYPEDATA_BORDER (&the_data);
    break;

  case OUTPUT:
    pipe_alignment (name, flag);
    pipe_size_offset_code (name, name, 0, NULL, flag);
    writebuf (func_body, "for (i = 0; i < sizeof (%s); i++) ", name);
    push_code_level();
    writebuf (func_body, "%s <<= 8;\n", name);
    writebuf (func_body, "%s |= i + 1;\n", name);
    pop_code_level();
    writebuf (func_body, "cptr = (char *)&(%s);\n", name);
    writebuf (func_body, "for (i = 0; i < sizeof (%s); i++) ", name);
    push_code_level();
    writebuf (func_body, "printf (\"%%d \", *(cptr+i));\n");
    pop_code_level();
    writebuf (func_body, "printf (\"0\\n\");\n");
    break;
  }
}

void pipe_pragmas (PreFlag flag)
{
  if (machine_info.char_info.size == DEFAULT_SIZE)
    pipe_pragma (&machine_info.char_info, flag);
  if (machine_info.short_info.size == DEFAULT_SIZE)
    pipe_pragma (&machine_info.short_info, flag);
  if (machine_info.int_info.size == DEFAULT_SIZE)
    pipe_pragma (&machine_info.int_info, flag);
  if (machine_info.long_info.size == DEFAULT_SIZE)
    pipe_pragma (&machine_info.long_info, flag);
}

void usit_new_function (void)
{
  static int label = 0;
  char name[256];

  sprintf (name, "usit_inf%d", label++);
  writebuf (UsitInferMain, "  %s();\n", name);

  gen_func_name_header ("int", name);
  writebuf (func_vars, "int i, j;\n");
  writebuf (func_vars, "char* cptr;\n");
}

void usit_end_function (void)
{
  gen_func_close();
}

void generate_code (UsitDecList *the_list, FILE *out)
{
  PreFlag flag = OUTPUT;
  UsitDecl *p;

  usit_new_function();
  pipe_pragmas (flag);
  usit_end_function();

  for (p = the_list->head; p; p = p->next) {
    usit_new_function();
    pipe_alignment (SYM_LOCAL_NAME (p->the_sym), flag);
    precode_sym (p->the_sym, p->the_sym, "", 0, -1, flag);
    usit_end_function();
  }
}

void chew_on_pipe (UsitDecList *the_list)
{
  PreFlag flag = INPUT;
  int alignment;
  UsitDecl *p;

  pipe_pragmas (flag);

  for (p = the_list->head; p; p = p->next) {
    alignment = pipe_alignment (SYM_LOCAL_NAME (p->the_sym), flag);
    precode_sym (p->the_sym, p->the_sym, "", 0, alignment, flag);
  }
}

void output_typedef (Symbol **sym, FILE *out)
{
  fprintf (out, "typedef ");
  declare_sym (0, *sym, out);
  fprintf (out, "\n");
}

void output_struct (Type **type, FILE *out)
{
  SpecifierTypeField specfield;

  specfield = TYPE_SPEC_TYPE (TYPE_ETYPE (*type));
  if (specfield & UNION_TYPE) {
    fprintf (out, "union ");
  }
  else if (specfield & STRUCT_TYPE) {
    fprintf (out, "struct ");
  }
  else {
    crash ("internal error in output_struct");
  }
  fprintf (out, "%s {\n", spec_type_to_string (TYPE_ETYPE (*type)));
  print_struct (0, TYPE_ETYPE (*type), out);
  fprintf (out, "};\n");
}

void ptab_check_and_set (Symbol **sym, int *dummy)
{
  SYM_NEXT (*sym) = NULL;
  set_all_sym_type_info (*sym);
}

void print_pragma (BucketInfo *info, FILE *out)
{
  fprintf (out, "#pragma %s(%d,%d,%d,%s)\n", info->name,
	   info->size, info->memsize, info->alignment, 
	   byteorder_to_string (info->byteorder));
}

void output_usc_file (FILE *out)
{
  print_pragma (&machine_info.char_info, out);
  print_pragma (&machine_info.short_info, out);
  print_pragma (&machine_info.int_info, out);
  print_pragma (&machine_info.long_info, out);
  fprintf (out, "\n");

  ptab (Type_Tab, ptab_check_and_set, NULL, 0);
  ptab (Type_Tab, output_typedef, out, 0);

/*  ptab (Struct_Defs, output_struct, out, 0);
  ptab (Union_Defs, output_struct, out, 0);
*/
}

void rm_temp (char *name)
{
#ifndef USE_TEMPC
  unlink (name);
#endif
}

void init_infer_main (void)
{
  UsitInferMain = makebuf (256);
  writebuf (UsitInferMain, 
	    "int main (void)\n"
	    "{\n");
}

void finish_infer_main (FILE *out)
{
  writebuf (UsitInferMain, "  return 0;\n}\n");
  fprintf (out, "%s", BUFFER_STRING (UsitInferMain));
  freebuf (UsitInferMain);
}


char *build_infer_code (UsitDecList *the_decls)
{
  FILE *cfile;
  char *filename;

  which_pass = 0;

  if (usit_output == USIT_OUT_USC_CODE) {
#ifdef USE_TEMPC
    filename = "tempc.c";
#else
    char *temp;
    temp = make_temp_file ();
    if (temp == NULL)
      crash ("Couldn't open temp file");
    filename = (char *)Malloc ((strlen (temp) + 3) * sizeof (char));
    sprintf (filename, "%s.c", temp);
    free (temp);
#endif
    if ((cfile = fopen (filename, "w")) == NULL) {
      crash ("Couldn't open temp file");
    }
  }
  else {
    filename = NULL;
    cfile = OutFile;
  }
 
  init_gen_utils (cfile);
  init_preout_file (cfile);
  init_infer_main ();
  declare_everything (the_decls, cfile);
  generate_code (the_decls, cfile);
  finish_infer_main (cfile);
  fclose (cfile);

  return filename;
}

char *compile_infer_code (char *c_filename)
{
  char *aout_filename;
  char buf[1024];
  int retvalue;

#ifdef USE_TEMPC
  aout_filename = "a.out";
#else
  aout_filename = make_temp_file();
#endif

  if (aout_filename == NULL)
    return NULL;

  sprintf (buf, "%s %s -o %s", native_compiler, c_filename, aout_filename);
  dprintf (0, "%s", buf);
  retvalue = system (buf);
  dprintf (0, "retvalue: %d", retvalue);
  rm_temp (c_filename);

  if (retvalue) {
    crash ("Error in input file caused preprocessor temp file "
	   "to not compile");
  }

  return aout_filename;
}

void run_preprocess (void)
{
  char *c_filename = NULL, *aout_filename = NULL;
  UsitDecList the_decls;
  
  init_dec_list (&the_decls);
     
  build_dec_list (&the_decls);
  print_dec_list (200, &the_decls);


  if (usit_input == USIT_IN_C_CODE) {
    c_filename = build_infer_code (&the_decls);
  }

  if (usit_output == USIT_OUT_USC_CODE) {
    if (usit_input == USIT_IN_C_CODE) {
      aout_filename = compile_infer_code (c_filename);
      if (aout_filename == NULL) {
	rm_temp (c_filename);
	crash ("Couldn't open temp file");
      }
      if ((pre_pipe = popen (aout_filename, "r")) == NULL)
	crash ("Can't execute preprocessor");
    }
    else {
      pre_pipe = UsitData;
    }
    dprintf (20, "Before pipe chewing:");
    print_type_table (300);
    chew_on_pipe (&the_decls);
    if (usit_input == USIT_IN_C_CODE) {
      pclose (pre_pipe);
      rm_temp (aout_filename);
    }
    else {
      fclose (pre_pipe);
    }
    which_pass = 1;
    dprintf (20, "Before output:");
    print_type_table (300);
    output_usc_file (OutFile);
  }

  /* already checked that make_validate is on only if usit_output == USC */

  if (make_validate) {
    init_validate();
    gen_usit_validate (&the_decls);
  }

  if (c_filename)
    Free (c_filename, 0);
#ifndef USE_TEMPC
  Free (aout_filename, 0);
#endif
}
